#!/usr/bin/env python
# -*-coding: utf-8 -*-
# ----------------------
# PySILLS.py
# Maximilian Beeskow
# 01.03.2021
# ----------------------
#
## MODULES
from modules import data
from modules import standard
from modules import sample
from modules import statistics
from modules import gui
import numpy as np
import matplotlib.pyplot as plt
import tkinter as tk
from tkinter import filedialog, Text
#
## IMPORT DATA
#
dataSRM = data.general()
CSTD = dataSRM.importSRM("../lib/NIST_610_Spandler.csv", ";", 0, 0)
#print("Test:\n", CSTD)
csdt = data.Data(filename="../lib/NIST_610_Spandler.csv")
cstdresults = csdt.import_data_to_pandas(delimiter=";", skip_header=None, skip_footer=0, names=["element", "concentration"])
#print(cstdresults)
dataInput = data.general()
IntensitiesData01 = dataInput.importData("../inputs/16nov14_a01.csv", ",", 3, 1)
#print(IntensitiesData01)
nRows = len(IntensitiesData01)
#print("Test (Intensities):\n", IntensitiesData01[1])
dataInput = data.general()
IntensitiesData02 = dataInput.importData("../inputs/16nov14_a02.csv", ",", 3, 1)
IntensitiesData03 = dataInput.importData("../inputs/16nov14_a03.csv", ",", 3, 1)
IntensitiesData04 = dataInput.importData("../inputs/16nov14_a04.csv", ",", 3, 1)
IntensitiesData05 = dataInput.importData("../inputs/16nov14_a05.csv", ",", 3, 1)
IntensitiesData06 = dataInput.importData("../inputs/16nov14_a06.csv", ",", 3, 1)
IntensitiesData07 = dataInput.importData("../inputs/16nov14_a07.csv", ",", 3, 1)
IntensitiesData08 = dataInput.importData("../inputs/16nov14_a08.csv", ",", 3, 1)
IntensitiesData09 = dataInput.importData("../inputs/16nov14_a09.csv", ",", 3, 1)
IntensitiesData10 = dataInput.importData("../inputs/16nov14_a10.csv", ",", 3, 1)
IntensitiesData11 = dataInput.importData("../inputs/16nov14_a11.csv", ",", 3, 1)
IntensitiesData12 = dataInput.importData("../inputs/16nov14_a12.csv", ",", 3, 1)
IntensitiesData13 = dataInput.importData("../inputs/16nov14_a13.csv", ",", 3, 1)
IntensitiesData14 = dataInput.importData("../inputs/16nov14_a14.csv", ",", 3, 1)
IntensitiesData15 = dataInput.importData("../inputs/16nov14_a15.csv", ",", 3, 1)
IntensitiesData16 = dataInput.importData("../inputs/16nov14_a16.csv", ",", 3, 1)
#
## SEGMENTATION
#
bg_time, sgn_time = data.Segmentation(IntensitiesData01).find_background()
bg_time_selected = data.Segmentation(IntensitiesData01).select_background(start_1=5.0, end_1=25.0, start_2=95.0, end_2=110.0)
sgn_time_selected = data.Segmentation(IntensitiesData01).select_signal(start=35.0, end=85.0)
dataSegmentation = data.segmentation(IntensitiesData01, False)
segments01, start01, end01 = dataSegmentation.segmentData()
print("Start 01:", start01, "End 01:", end01)
dataSegmentation = data.segmentation(IntensitiesData02, False)
segments02, start02, end02 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData03, False)
segments03, start03, end03 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData04, False)
segments04, start04, end04 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData05, False)
segments05, start05, end05 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData06, False)
segments06, start06, end06 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData07, False)
segments07, start07, end07 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData08, False)
segments08, start08, end08 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData09, False)
segments09, start09, end09 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData10, False)
segments10, start10, end10 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData11, False)
segments11, start11, end11 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData12, False)
segments12, start12, end12 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData13, False)
segments13, start13, end13 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData14, False)
segments14, start14, end14 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData15, False)
segments15, start15, end15 = dataSegmentation.segmentData()
dataSegmentation = data.segmentation(IntensitiesData16, False)
segments16, start16, end16 = dataSegmentation.segmentData()
#
## SENSITIVITIES STD
#
data = standard.standardAnalysis()
refData, xi01, Ccalc01, C_STD = data.calculateSensitivities(CSTD, IntensitiesData01, start01, end01, "Si", False)
refData, xi02, Ccalc02, C_STD = data.calculateSensitivities(CSTD, IntensitiesData02, start02, end02, "Si", False)
refData, xi15, Ccalc15, C_STD = data.calculateSensitivities(CSTD, IntensitiesData15, start15, end15, "Si", False)
refData, xi16, Ccalc16, C_STD = data.calculateSensitivities(CSTD, IntensitiesData16, start16, end16, "Si", False)
#
xi_01, isotope_list_01 = data.calculate_xi(IntensitiesData01, CSTD, "Si", start01, end01)
xi_02, isotope_list_02 = data.calculate_xi(IntensitiesData02, CSTD, "Si", start02, end02)
xi_15, isotope_list_15 = data.calculate_xi(IntensitiesData15, CSTD, "Si", start15, end15)
xi_16, isotope_list_16 = data.calculate_xi(IntensitiesData16, CSTD, "Si", start16, end16)
#
data = statistics.statisticalAnalysis([xi01, xi02, xi15, xi16])
analyzedSensitivities = data.analyzeSensitivities(reference="Si")
#
data = statistics.statisticalAnalysis([xi_01, xi_02, xi_15, xi_16])
xi_analyzed, df_xi = data.analyze_xi(reference="Si", isotopes=isotope_list_01)
#
## SAMPLE CONCENTRATIONS
#
exampleConcentration = 260774
data = sample.sampleAnalysis()
Csamp03, refData03 = data.calcConcentration(IntensitiesData03, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
C_sample_03, df_c_03 = data.calculate_concentrations(IntensitiesData03, xi=df_xi, start=start03, stop=end03, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
C_sample_04, df_c_04 = data.calculate_concentrations(IntensitiesData04, xi=df_xi, start=start04, stop=end04, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
C_sample_05, df_c_05 = data.calculate_concentrations(IntensitiesData05, xi=df_xi, start=start05, stop=end05, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
C_sample_06, df_c_06 = data.calculate_concentrations(IntensitiesData06, xi=df_xi, start=start06, stop=end06, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
C_sample_07, df_c_07 = data.calculate_concentrations(IntensitiesData07, xi=df_xi, start=start07, stop=end07, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
C_sample_08, df_c_08 = data.calculate_concentrations(IntensitiesData08, xi=df_xi, start=start08, stop=end08, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
C_sample_09, df_c_09 = data.calculate_concentrations(IntensitiesData09, xi=df_xi, start=start09, stop=end09, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
C_sample_10, df_c_10 = data.calculate_concentrations(IntensitiesData10, xi=df_xi, start=start10, stop=end10, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
C_sample_11, df_c_11 = data.calculate_concentrations(IntensitiesData11, xi=df_xi, start=start11, stop=end11, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
C_sample_12, df_c_12 = data.calculate_concentrations(IntensitiesData12, xi=df_xi, start=start12, stop=end12, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
C_sample_13, df_c_13 = data.calculate_concentrations(IntensitiesData13, xi=df_xi, start=start13, stop=end13, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
C_sample_14, df_c_14 = data.calculate_concentrations(IntensitiesData14, xi=df_xi, start=start14, stop=end14, C_ref=exampleConcentration, reference="Si", isotopes=isotope_list_01)
Csamp04, refData04 = data.calcConcentration(IntensitiesData04, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
Csamp05, refData05 = data.calcConcentration(IntensitiesData05, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
Csamp06, refData06 = data.calcConcentration(IntensitiesData06, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
Csamp07, refData07 = data.calcConcentration(IntensitiesData07, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
Csamp08, refData08 = data.calcConcentration(IntensitiesData08, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
Csamp09, refData09 = data.calcConcentration(IntensitiesData09, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
Csamp10, refData10 = data.calcConcentration(IntensitiesData10, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
Csamp11, refData11 = data.calcConcentration(IntensitiesData11, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
Csamp12, refData12 = data.calcConcentration(IntensitiesData12, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
Csamp13, refData13 = data.calcConcentration(IntensitiesData13, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
Csamp14, refData14 = data.calcConcentration(IntensitiesData14, exampleConcentration, analyzedSensitivities, start03, end03, "Si29", False)
#
#df_i, df_c = data.create_dataframe([IntensitiesData01, IntensitiesData02, IntensitiesData15, IntensitiesData16], [CSTD], isotope_list_01)
rsf_03, df_i_03, df_c_03 = data.calculate_rsf(IntensitiesData03, IntensitiesData01, C_sample_03, CSTD, "Si", isotope_list_01)
#
data = statistics.statisticalAnalysis([Csamp03, Csamp04, Csamp05, Csamp06, Csamp06, Csamp07, Csamp08, Csamp09, Csamp10, Csamp11, Csamp12, Csamp13, Csamp14])
analyzedConcentrations = data.analyzeConcentrations()
#
## RSF
data = sample.sampleAnalysis()
RSF03 = data.calcRSF([IntensitiesData01, IntensitiesData03], [C_STD, Csamp03], analyzedSensitivities, [start03, end03], "Si29")
#
## LOD
data = sample.sampleAnalysis()
#LOD3 = data.calculateLOD(RSF03, [IntensitiesData01, IntensitiesData03], C_STD, start03)
#


#!/usr/bin/env python
# -*-coding: utf-8 -*-

#-----------------------------------------------

# Name:		spike_elimination.py
# Author:	Maximilian A. Beeskow
# Version:	v1.0.64
# Date:		07.02.2025

#-----------------------------------------------

## MODULES
import numpy as np
import scipy.stats as stats
from math import sqrt
from collections import defaultdict
import pandas as pd

## CLASSES
class OutlierDetection:
    def __init__(self, raw_data, alpha, threshold, isotope, dataset_complete):
        self.raw_data = raw_data
        self.alpha = alpha
        self.threshold = threshold
        self.smoothed_data = []
        self.spike_indices = []
        self.isotope = isotope
        self.dataset_complete = dataset_complete

    def calculate_whisker_statistcs(self, helper_values, init=True):
        helper_dataset = helper_values
        val_quartile_first = np.quantile(helper_dataset, 0.25)
        val_quartile_third = np.quantile(helper_dataset, 0.75)
        val_iqr = val_quartile_third - val_quartile_first
        factor = abs(self.alpha + 1)
        val_lower_limit = val_quartile_first - factor*val_iqr
        val_upper_limit = val_quartile_third + factor*val_iqr

        if init == True:
            lower_outliers = [index for index, value in enumerate(helper_dataset) if value < val_lower_limit and
                              value > self.threshold]
            upper_outliers = [index for index, value in enumerate(helper_dataset) if value > val_upper_limit and
                              value > self.threshold]

            outlier_indices = lower_outliers
            outlier_indices.extend(upper_outliers)
            outlier_indices.sort()
        else:
            value_poi = helper_dataset[0]
            print(helper_dataset, helper_dataset[0:])
            val_mean = np.mean(helper_dataset[0:])
            values_sp = helper_dataset[0:]

    def find_outlier(self):
        val_quartile_first = np.quantile(self.dataset_complete, 0.25)
        val_quartile_third = np.quantile(self.dataset_complete, 0.75)
        val_iqr = val_quartile_third - val_quartile_first
        factor = abs(self.alpha + 1)
        val_lower_limit = val_quartile_first - factor*val_iqr
        val_upper_limit = val_quartile_third + factor*val_iqr

        lower_outliers = [index for index, value in enumerate(self.dataset_complete) if value < val_lower_limit and
                          value > self.threshold]
        upper_outliers = [index for index, value in enumerate(self.dataset_complete) if value > val_upper_limit and
                          value > self.threshold]

        data_smoothed = self.dataset_complete.copy()
        outlier_indices = lower_outliers
        outlier_indices.extend(upper_outliers)
        outlier_indices.sort()

        for index in outlier_indices:
            helper_values = self.determine_surrounded_values(var_index=index, stepsize=4)

            if index in lower_outliers:
                corrected_value = np.mean(helper_values["SP"])
            elif index in upper_outliers:
                corrected_value = np.mean(helper_values["SP"])

            data_smoothed[index] = corrected_value

        return data_smoothed, outlier_indices
    def determine_surrounded_values(self, var_index, stepsize=3):
        # POI   - Point Of Interest
        # SP    - Surrounding Points
        helper_values = {"POI": 0, "SP": [], "All": []}
        val_poi = self.dataset_complete[var_index]
        helper_values["POI"] = val_poi
        helper_values["All"].append(val_poi)

        for step in range(1, stepsize):
            step_before = var_index - step
            step_after = var_index + step
            if step_before >= 0:
                helper_values["SP"].append(self.dataset_complete[step_before])
                helper_values["All"].append(self.dataset_complete[step_before])
            if step_after < len(self.dataset_complete):
                helper_values["SP"].append(self.dataset_complete[step_after])
                helper_values["All"].append(self.dataset_complete[step_after])

        return helper_values

    def extract_surrounding_values(self, index_spike, steps):
        helper_values = {"POI": 0, "SP": [], "All": []}
        val_poi = self.dataset_complete[index_spike]
        helper_values["POI"] = val_poi

        index_start = max(0, index_spike - steps)
        index_end = min(len(self.dataset_complete), index_spike + (steps + 1))

        helper_values["All"].append(self.dataset_complete[index_start:index_end])
        helper_values["SP"].append((self.dataset_complete[index_start:index_spike] +
                                    self.dataset_complete[index_spike + 1:index_end]))

        return helper_values

    def calculate_grubbs_critical_value(self, size):
        t_dist = stats.t.ppf(1 - self.alpha/(2*size), size - 2)
        numerator = (size - 1)*np.sqrt(np.square(t_dist))
        denominator = np.sqrt(size)*np.sqrt(size - 2 + np.square(t_dist))
        critical_value = numerator/denominator

        return critical_value

class GrubbsTestSILLS:
    def __init__(self, raw_data, alpha, threshold, start_index, dataset_complete):
        self.raw_data = raw_data
        self.alpha = alpha
        self.threshold = threshold
        #self.val_check_7 = 2.097
        #self.val_check_9 = 2.323
        self.smoothed_data = []
        self.spike_indices = []
        self.start_index = start_index
        self.dataset_complete = dataset_complete

    def calculate_grubbs_critical(self, N, alpha=0.01):
        df = N - 2
        t_val = stats.t.ppf(1 - (alpha/(2*N)), df)
        grubbs_val = (N - 1)/(N**0.5)*((t_val**2)/(df + t_val**2))**0.5
        return grubbs_val

    def outlier_elimination_zsore(self): # outlier_elimination_zsore
        data = self.raw_data
        z_scores = np.abs(stats.zscore(data))
        zscore_threshold = 2
        helper_data = {}

        for index, value in enumerate(self.raw_data):
            if value > self.threshold:
                if z_scores[index] > zscore_threshold:
                    self.spike_indices.append(index + self.start_index)
                    surrounding_6 = self.extract_surrounding_values(index_spike=index, steps=4)
                    mean_5 = np.mean(surrounding_6["SP"])
                    helper_data[index + self.start_index] = mean_5

        for index, value in enumerate(self.dataset_complete):
            if index in self.spike_indices:
                self.smoothed_data.append(helper_data[index])
            else:
                self.smoothed_data.append(value)

        return self.smoothed_data, self.spike_indices

    def determine_outlier(self):
        N_low = 7
        N_high = 9
        G_comp_7 = round(self.calculate_grubbs_critical(N=N_low), 3)
        G_comp_9 = round(self.calculate_grubbs_critical(N=N_high), 3)
        helper_data = {}

        for index, value in enumerate(self.raw_data):
            if value > self.threshold:
                surrounding_7 = self.extract_surrounding_values(index_spike=index, steps=N_low)
                surrounding_9 = self.extract_surrounding_values(index_spike=index, steps=N_high)

                mean_6 = np.mean(surrounding_7["SP"])
                mean_7 = np.mean(surrounding_7["All"])
                std_7 = np.std(surrounding_7["All"], ddof=1)
                mean_9 = np.mean(surrounding_9["All"])
                std_9 = np.std(surrounding_9["All"], ddof=1)

                if np.abs(value - mean_7)/std_7 > G_comp_7 and np.abs(value - mean_9)/std_9 > G_comp_9:
                    val_corr = mean_6
                    self.spike_indices.append(index + self.start_index)
                    helper_data[index + self.start_index] = val_corr

        for index, value in enumerate(self.dataset_complete):
            if index in self.spike_indices:
                self.smoothed_data.append(helper_data[index])
            else:
                self.smoothed_data.append(value)

        return self.smoothed_data, self.spike_indices

    def determine_outlier2(self):
        helper_data = {}
        for index, value in enumerate(self.raw_data):
            if value >= self.threshold:
                surrounding_3 = self.determine_surrounded_values(index=index, stepsize=3)
                surrounding_4 = self.determine_surrounded_values(index=index, stepsize=4)

                mean_3 = np.mean(surrounding_3["All"])
                std_3 = np.std(surrounding_3["All"], ddof=1)
                mean_4 = np.mean(surrounding_4["All"])
                std_4 = np.std(surrounding_4["All"], ddof=1)

                if std_3 > 0:
                    val_poi_3 = round(abs(value - mean_3)/std_3, 3)
                else:
                    val_poi_3 = np.nan

                if std_4 > 0:
                    val_poi_4 = round(abs(value - mean_4)/std_4, 3)
                else:
                    val_poi_4 = np.nan

                val_crit_3 = round(self.calculate_grubbs_critical_value(size=len(surrounding_3["All"])), 3)
                val_crit_4 = round(self.calculate_grubbs_critical_value(size=len(surrounding_4["All"])), 3)

                if val_poi_3 > val_crit_3 and val_poi_4 > val_crit_4:
                    val_corr = np.mean(surrounding_3["SP"])
                    self.spike_indices.append(index + self.start_index)
                    helper_data[index + self.start_index] = val_corr

        for index, value in enumerate(self.dataset_complete):
            if index in self.spike_indices:
                self.smoothed_data.append(helper_data[index])
            else:
                self.smoothed_data.append(value)

        return self.smoothed_data, self.spike_indices

    def extract_surrounding_values(self, index_spike, steps):
        helper_values = {"POI": 0, "SP": [], "All": []}
        val_poi = self.dataset_complete[index_spike]
        helper_values["POI"] = val_poi

        index_start = max(0, index_spike - steps)
        index_end = min(len(self.dataset_complete), index_spike + (steps + 1))

        helper_values["All"].append(self.dataset_complete[index_start:index_end])
        helper_values["SP"].append((self.dataset_complete[index_start:index_spike] +
                                    self.dataset_complete[index_spike + 1:index_end]))

        return helper_values

    def determine_surrounded_values(self, index, stepsize=4):
        helper_values = {"POI": 0, "SP": [], "All": []}
        val_poi = self.raw_data[index]
        helper_values["POI"] = val_poi
        helper_values["All"].append(val_poi)

        for step in range(1, stepsize):
            step_before = index - step
            step_after = index + step
            if step_before >= 0:
                helper_values["SP"].append(self.raw_data[step_before])
                helper_values["All"].append(self.raw_data[step_before])
            if step_after < len(self.raw_data):
                helper_values["SP"].append(self.raw_data[step_after])
                helper_values["All"].append(self.raw_data[step_after])

        return helper_values

    def calculate_grubbs_critical_value(self, size):
        t_dist = stats.t.ppf(1 - self.alpha/(2*size), size - 2)
        numerator = (size - 1)*np.sqrt(np.square(t_dist))
        denominator = np.sqrt(size)*np.sqrt(size - 2 + np.square(t_dist))
        critical_value = numerator/denominator

        return critical_value

class GrubbsTest:
    def __init__(self, raw_data, alpha, threshold, start_index, dataset_complete):
        self.raw_data = raw_data
        self.alpha = alpha
        self.threshold = threshold
        self.smoothed_data = []
        self.spike_indices = []
        self.start_index = start_index
        self.dataset_complete = dataset_complete

    def determine_outlier(self):
        helper_data = {}
        for index, value in enumerate(self.raw_data):
            if value >= self.threshold:
                surrounding_3 = self.determine_surrounded_values(index=index, stepsize=3)
                surrounding_4 = self.determine_surrounded_values(index=index, stepsize=4)
                val_poi_3, index_poi_3 = self.calculate_grubbs_value()
                val_poi_3 = round(val_poi_3, 3)
                val_crit_3 = round(self.calculate_grubbs_critical_value(size=len(surrounding_3["All"])), 3)
                val_poi_4, index_poi_4 = self.calculate_grubbs_value()
                val_poi_4 = round(val_poi_4, 3)
                val_crit_4 = round(self.calculate_grubbs_critical_value(size=len(surrounding_4["All"])), 3)

                if val_poi_3 > val_crit_3 and val_poi_4 > val_crit_4:
                    val_corr = np.mean(surrounding_3["SP"])
                    self.spike_indices.append(index + self.start_index)
                    helper_data[index + self.start_index] = val_corr

        for index, value in enumerate(self.dataset_complete):
            if index in self.spike_indices:
                self.smoothed_data.append(helper_data[index])
            else:
                self.smoothed_data.append(value)

        return self.smoothed_data, self.spike_indices

    def determine_surrounded_values(self, index, stepsize=4):
        helper_values = {"POI": 0, "SP": [], "All": []}
        val_poi = self.raw_data[index]
        helper_values["POI"] = val_poi
        helper_values["All"].append(val_poi)

        for step in range(1, stepsize):
            step_before = index - step
            step_after = index + step
            if step_before >= 0:
                helper_values["SP"].append(self.raw_data[step_before])
                helper_values["All"].append(self.raw_data[step_before])
            if step_after < len(self.raw_data):
                helper_values["SP"].append(self.raw_data[step_after])
                helper_values["All"].append(self.raw_data[step_after])

        return helper_values

    def calculate_grubbs_critical_value(self, size):
        t_dist = stats.t.ppf(1 - self.alpha/(2*size), size - 2)
        numerator = (size - 1)*np.sqrt(np.square(t_dist))
        denominator = np.sqrt(size)*np.sqrt(size - 2 + np.square(t_dist))
        critical_value = numerator/denominator

        return critical_value

    def calculate_grubbs_value(self):
        std_dev = np.std(self.raw_data, ddof=1)
        avg_y = np.mean(self.raw_data)
        abs_val_minus_avg = abs(np.array(self.raw_data) - avg_y)
        max_of_deviations = max(abs_val_minus_avg)
        max_ind = np.argmax(abs_val_minus_avg)
        Gcal = max_of_deviations/std_dev

        return Gcal, max_ind


__all__ = ['two_sided_test_indices',
           'two_sided_test_outliers',
           'TwoSidedGrubbsTest',
           'OutputType']


DEFAULT_ALPHA = 0.95


# Test output types
class OutputType:
    DATA = 0  # Output data without outliers
    OUTLIERS = 1  # Output outliers
    INDICES = 2  # Output outlier indices


class GrubbsTestAlternative(object):
    def __init__(self, data):
        self.original_data = data

    def _copy_data(self):
        if isinstance(self.original_data, np.ndarray):
            return self.original_data
        elif pd is not None and isinstance(self.original_data, pd.Series):
            return self.original_data
        elif isinstance(self.original_data, list):
            return np.array(self.original_data)
        else:
            raise TypeError('Unsupported data format')

    def _delete_item(self, data, index):
        if pd is not None and isinstance(data, pd.Series):
            return data.drop(index)
        elif isinstance(data, np.ndarray):
            return np.delete(data, index)
        else:
            raise TypeError('Unsupported data format')

    def _get_indices(self, values):
        last_seen = defaultdict(lambda: 0)
        data = list(self.original_data)
        indices = list()
        for value in values:
            start = last_seen[value]
            index = data.index(value, start)
            indices.append(index)
            last_seen[value] = index + 1
        return indices

    def _get_g_test(self, data, alpha):
        """Compute a significant value score following these steps, being alpha
        the requested significance level:

        1. Find the upper critical value of the t-distribution with n-2
           degrees of freedom and a significance level of alpha/2n
           (for two-sided tests) or alpha/n (for one-sided tests).

        2. Use this t value to find the score with the following formula:
           ((n-1) / sqrt(n)) * (sqrt(t**2 / (n-2 + t**2)))

        :param numpy.array data: data set
        :param float alpha: significance level
        :return: G_test score
        """
        n = len(data)
        significance_level = self._get_t_significance_level(alpha, n)
        t = stats.t.isf(significance_level, n-2)
        return ((n-1) / sqrt(n)) * (sqrt(t**2 / (n-2 + t**2)))

    def _test_once(self, data, alpha):
        """Perform one iteration of the Smirnov-Grubbs test.

        :param numpy.array data: data set
        :param float alpha: significance level
        :return: the index of the outlier if one if found; None otherwise
        """
        target_index, value = self._target(data)
        var_std = data.std()
        if var_std > 0:
            g = value/var_std
        else:
            g = 0
        g_test = self._get_g_test(data, alpha)
        return target_index if g > g_test else None

    def run(self, alpha=DEFAULT_ALPHA, output_type=OutputType.DATA):
        """Run the Smirnov-Grubbs test to remove outliers in the given data set.

        :param float alpha: significance level
        :param int output_type: test output type (from OutputType class values)
        :return: depending on the value of output_type, the data set without
        outliers (DATA), the outliers themselves (OUTLIERS) or the indices of
        the outliers in the original data set (INDICES)
        """
        data = self._copy_data()
        outliers = list()

        while True:
            outlier_index = self._test_once(data, alpha)
            if outlier_index is None:
                break
            outlier = data[outlier_index]
            outliers.append(outlier)
            data = self._delete_item(data, outlier_index)

        return_value = data
        if output_type == OutputType.OUTLIERS:
            return_value = outliers
        elif output_type == OutputType.INDICES:
            return_value = self._get_indices(outliers)
        return return_value

    def _target(self, data):
        raise NotImplementedError

    def _get_t_significance_level(self, alpha):
        raise NotImplementedError

class TwoSidedGrubbsTest(GrubbsTestAlternative):
    def _target(self, data):
        """Compute the index of the farthest value from the sample mean and its
        distance.

        :param numpy.array data: data set
        :return int, float: the index of the element and its distance to the
        mean
        """
        relative_values = abs(data - data.mean())
        index = relative_values.argmax()
        value = relative_values[index]
        return index, value

    def _get_t_significance_level(self, alpha, n):
        return alpha/(2*n)

def _test(test_class, data, alpha, output_type):
    return test_class(data).run(alpha, output_type=output_type)

def _two_sided_test(data, alpha, output_type):
    return _test(TwoSidedGrubbsTest, data, alpha, output_type)

def two_sided_test_indices(data, alpha=DEFAULT_ALPHA):
    #alpha = 1 - alpha
    return _two_sided_test(data, alpha, OutputType.INDICES)

def two_sided_test_outliers(data, alpha=DEFAULT_ALPHA):
    return _two_sided_test(data, alpha, OutputType.OUTLIERS)


# -*- coding: utf-8 -*-
# Author: Yiping Liu
# Description: This script calculates the average of a list of numbers.
# Version: 1.0
# Last Modified: May 7, 2023


from sympy import *
import pprint

pp = pprint.PrettyPrinter(width=41, compact=True)

x, y, z = symbols('x,y,z')

def newton_raphson_general(x_n, y_n, f_func, g_func, molalities, max_iterators):
    partial_fx = diff(f_func, x)
    partial_fy = diff(f_func, y)
    partial_gx = diff(g_func, x)
    partial_gy = diff(g_func, y)
    tolerance = 10 ** (-6)
    iterate_number = 0
    while iterate_number < max_iterators:
        print("x_{n}: {x_n}; y_{n}: {y_n}".format(n=iterate_number, x_n=x_n, y_n=y_n))
        iterate_number += 1
        partial_fx = partial_fx.subs({x: x_n, y: y_n}).evalf()
        partial_fy = partial_fy.subs({x: x_n, y: y_n}).evalf()
        partial_gx = partial_gx.subs({x: x_n, y: y_n}).evalf()
        partial_gy = partial_gy.subs({x: x_n, y: y_n}).evalf()
        det_value = partial_fx * partial_gy - partial_fy * partial_gx
        f_value = f_func.subs({x: x_n, y: y_n}).evalf()
        g_value = g_func.subs({x: x_n, y: y_n}).evalf()
        x_n_plus = x_n - (1 / det_value) * (partial_gy * f_value - partial_fy * g_value)
        y_n_plus = y_n - (1 / det_value) * (-partial_gx * f_value + partial_fx * g_value)
        if abs((x_n_plus - x_n) / x_n) < tolerance and abs((y_n_plus - y_n) / y_n) < tolerance:
            for key, value in molalities.items():
                molalities[key] = value.subs({x: x_n, y: y_n})
            return molalities, y_n
        x_n = x_n_plus
        y_n = y_n_plus


def newton_raphson_three(funcs, x_n, y_n, z_n, molalities, max_iterators):
    f_func = funcs[0]
    g_func = funcs[1]
    h_func = funcs[2]
    a1 = diff(f_func, x)  # partial_fx
    a2 = diff(f_func, y)  # partial_fy
    a3 = diff(f_func, z)  # partial_fz

    b1 = diff(g_func, x)  # partial_gx
    b2 = diff(g_func, y)  # partial_gy
    b3 = diff(g_func, z)  # partial_gz

    c1 = diff(h_func, x)  # partial_hx
    c2 = diff(h_func, y)  # partial_hy
    c3 = diff(h_func, z)  # partial_hz

    tolerance = 10 ** (-6)
    iterate_number = 0
    while iterate_number < max_iterators:
        print("x_{n}: {x_n}; y_{n}: {y_n}; z_{n}: {z_n}".format(n=iterate_number, x_n=x_n, y_n=y_n, z_n=z_n))
        iterate_number += 1
        a1 = a1.subs({x: x_n, y: y_n, z: z_n}).evalf()
        a2 = a2.subs({x: x_n, y: y_n, z: z_n}).evalf()
        a3 = a3.subs({x: x_n, y: y_n, z: z_n}).evalf()
        b1 = b1.subs({x: x_n, y: y_n, z: z_n}).evalf()
        b2 = b2.subs({x: x_n, y: y_n, z: z_n}).evalf()
        b3 = b3.subs({x: x_n, y: y_n, z: z_n}).evalf()
        c1 = c1.subs({x: x_n, y: y_n, z: z_n}).evalf()
        c2 = c2.subs({x: x_n, y: y_n, z: z_n}).evalf()
        c3 = c3.subs({x: x_n, y: y_n, z: z_n}).evalf()

        det_value = a1 * b2 * c3 - a1 * b3 * c2 - a2 * b1 * c3 + a2 * b3 * c1 + a3 * b1 * c2 - a3 * b2 * c1

        f_value = f_func.subs({x: x_n, y: y_n, z: z_n}).evalf()
        g_value = g_func.subs({x: x_n, y: y_n, z: z_n}).evalf()
        h_value = h_func.subs({x: x_n, y: y_n, z: z_n}).evalf()

        x_n_plus = x_n - (1 / det_value) * (
                (b2 * c3 - b3 * c2) * f_value + (-a2 * c3 + a3 * c2) * g_value + (a2 * b3 - a3 * b2) * h_value)
        y_n_plus = y_n - (1 / det_value) * (
                (-b1 * c3 + b3 * c1) * f_value + (a1 * c3 - a3 * c1) * g_value + (-a1 * b3 + a3 * b1) * h_value)
        z_n_plus = z_n - (1 / det_value) * (
                (b1 * c2 - b2 * c1) * f_value + (-a1 * c2 + a2 * c1) * g_value + (a1 * b2 - a2 * b1) * h_value)
        if abs((x_n_plus - x_n) / x_n) < tolerance and abs((y_n_plus - y_n) / y_n) < tolerance and abs(
                (z_n_plus - z_n) / z_n) < tolerance:
            for key, value in molalities.items():
                try:
                    molalities[key] = value.subs({x: x_n, y: y_n, z: z_n})
                except AttributeError:
                    molalities[key] = value
            return molalities, z_n
        x_n = x_n_plus
        y_n = y_n_plus
        z_n = z_n_plus


#!/usr/bin/env python
# -*-coding: utf-8 -*-

#-----------------------------------------------

# Name:		mineral_analysis.py
# Author:	Maximilian A. Beeskow
# Version:	1.0
# Date:		31.03.2023

#-----------------------------------------------

## MODULES
import math
import re, datetime
import numpy as np
from modules.data import Data

class MineralAnalysisCalculations:
    #
    def __init__(self, var_filetype, var_datatype, var_isotopes, var_results):
        pass

class MineralAnalysis:
    #
    def __init__(self, container_measurements, container_lists, container_files, container_var, srm_data, var_filetype,
                 var_datatype, var_is, xi_std_time):
        self.container_measurements = container_measurements
        self.container_lists = container_lists
        self.container_files = container_files
        self.container_var = container_var
        self.srm_data = srm_data
        self.var_filetype = var_filetype
        self.var_datatype = var_datatype
        self.var_is = var_is
        self.xi_std_time = xi_std_time
        #
        self.srm_isotopes = {}
        for isotope in self.container_lists["ISOTOPES"]:
            self.srm_isotopes[isotope] = {}
            var_srm = self.container_var["SRM"][isotope].get()
            key_element = re.search("(\D+)(\d+)", isotope)
            var_element = key_element.group(1)
            self.srm_isotopes[isotope]["SRM"] = var_srm
            if var_element in self.srm_data[var_srm]:
                self.srm_isotopes[isotope]["Concentration"] = self.srm_data[var_srm][var_element]
            else:
                self.srm_isotopes[isotope]["Concentration"] = 0.0
        #
        self.isotope_by_srm = {}
        for key, value in self.container_files["SRM"].items():
            var_srm = value.get()
            #
            if var_srm not in self.isotope_by_srm:
                self.isotope_by_srm[var_srm] = []
            #
            self.isotope_by_srm[var_srm].append(key)
        #
        category_results = ["Intensity BG", "Intensity SIG CORR", "Sensitivity", "Concentration", "RSF", "S", "LOD"]
        self.container_lists["Assemblages"] = {}
        for file_smpl in self.container_lists["SMPL"]["Long"]:
            parts = file_smpl.split("/")
            file_smpl_short = parts[-1]
            #
            var_id =self.container_var["SMPL"][file_smpl]["ID"].get()
            if var_id not in self.container_lists["Assemblages"]:
                self.container_lists["Assemblages"][var_id] = {}
                self.container_lists["Assemblages"][var_id][file_smpl_short] = {}
                for category in category_results:
                    self.container_lists["Assemblages"][var_id][file_smpl_short][category] = {}
                    for isotope in self.container_lists["ISOTOPES"]:
                        self.container_lists["Assemblages"][var_id][file_smpl_short][category][isotope] = 0
        #
        self.list_assemblages = list(self.container_lists["Assemblages"].keys())
        #
        self.container_results = {"STD": {"ALL": {}, "FILE": {}}, "SMPL": {"ALL": {}, "FILE": {}}}
        #
        for file_std in self.container_lists["STD"]["Short"]:
            self.container_results["STD"]["FILE"][file_std] = {}
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            self.container_results["SMPL"]["FILE"][file_smpl] = {}
        #
        for var_id, var_files in self.container_lists["ID Files"].items():
            self.container_results["SMPL"][var_id] = {"ALL": {}}
            for var_file in var_files:
                self.container_results["SMPL"][var_id][var_file] = {}

        # for key1, values in self.container_results.items():
        #     print(key1)
        #     for key2, value in values.items():
        #         print(key2, value)

    #
    def get_intensity_single(self):
        helper = {}
        #
        for isotope in self.container_lists["ISOTOPES"]:
            helper[isotope] = []
            #
            if self.data_section == "BG":
                intensity_i = np.nanmean(
                self.container_measurements["SELECTED"][self.filename_single][self.var_datatype][isotope]["BG"])
            elif self.data_section == "SIG":
                intensity_i = np.nanmean(
                self.container_measurements["SELECTED"][self.filename_single][self.var_datatype][isotope]["SIG"])
            elif self.data_section == "SIG CORR":
                intensity_i_bg = np.nanmean(
                    self.container_measurements["SELECTED"][self.filename_single][self.var_datatype][isotope]["BG"])
                intensity_i_mat = np.nanmean(
                    self.container_measurements["SELECTED"][self.filename_single][self.var_datatype][isotope]["SIG"])
                #
                if np.nanmean(intensity_i_mat) >= np.nanmean(intensity_i_bg):
                    intensity_i = intensity_i_mat - np.nanmean(intensity_i_bg)
                else:
                    intensity_i = 0.0
            #
            helper[isotope] = intensity_i
        #
        # print("TESTING: INTENSITY")
        # for key, value in helper.items():
        #     print(key, value)
        #
        return helper
    #
    def get_intensity_ratio_single(self):
        helper = {}
        #
        for isotope in self.container_lists["ISOTOPES"]:
            helper[isotope] = []
            #
            intensity_i = self.results_single["Intensity"][isotope]
            intensity_is = self.results_single["Intensity"][self.var_is]
            intensity_ratio_i = intensity_i/intensity_is
            #
            helper[isotope] = intensity_ratio_i
        #
        return helper
    #
    def get_sensitivity_single(self):
        helper = {}
        #
        if self.var_filetype == "STD":
            concentration_is = self.srm_isotopes[self.var_is]["Concentration"]
            #
            for isotope in self.container_lists["ISOTOPES"]:
                helper[isotope] = []
                #
                concentration_i = self.srm_isotopes[isotope]["Concentration"]
                #
                if concentration_i > 0:
                    sensitivity_i = self.results_single["Intensity Ratio"][isotope]*(concentration_is/concentration_i)
                else:
                    sensitivity_i = 0.0
                #
                helper[isotope] = sensitivity_i
        #
        elif self.var_filetype == "SMPL":
            self.extract_data_times()
            helper_sensitivity_std = {}
            xi_opt = {}
            #
            for index, file_std in enumerate(self.container_lists["STD"]["Short"]):
                self.xi_std_time[file_std] = {}
                #
                intensity_is_bg = np.mean(
                    self.container_measurements["SELECTED"][file_std][self.var_datatype][self.var_is]["BG"])
                intensity_is_mat = np.mean(
                    self.container_measurements["SELECTED"][file_std][self.var_datatype][self.var_is]["SIG"])
                #
                if np.mean(intensity_is_mat) >= np.mean(intensity_is_bg):
                    intensity_is = intensity_is_mat - np.mean(intensity_is_bg)
                else:
                    intensity_is = 0.0
                #
                concentration_is = self.srm_isotopes[self.var_is]["Concentration"]
                #
                for isotope, value in self.container_measurements["SELECTED"][file_std][self.var_datatype].items():
                    if isotope not in helper_sensitivity_std:
                        xi_opt[isotope] = []
                        helper_sensitivity_std[isotope] = []
                    intensity_i_bg = np.mean(value["BG"])
                    intensity_i_mat = np.mean(value["SIG"])
                    if np.mean(intensity_is_mat) >= np.mean(intensity_is_bg):
                        intensity_i = intensity_i_mat - np.mean(intensity_i_bg)
                    else:
                        intensity_i = 0.0
                        # intensity_i = intensity_i_bg - np.mean(intensity_i_mat)
                    #
                    concentration_i = self.srm_isotopes[isotope]["Concentration"]
                    value = (intensity_i*concentration_is)/(intensity_is*concentration_i)
                    #
                    helper_sensitivity_std[isotope].append(value)
                    #
                    if isotope not in self.xi_std_time[file_std]:
                        self.xi_std_time[file_std][isotope] = [
                            self.std_times[file_std]["Delta"], np.mean(value.tolist())]
            for isotope in self.container_lists["ISOTOPES"]:
                xi_regr = self.calculate_regression(
                    data=self.xi_std_time, isotope=isotope, file_data=self.container_lists["STD"]["Short"])
                xi_opt[isotope].extend(xi_regr)
            #
            for isotope in self.container_lists["ISOTOPES"]:
                helper[isotope] = []
                value = xi_opt[isotope][0]*self.smpl_times[self.filename_single]["Delta"] + xi_opt[isotope][1]
                if value >= 0:
                    helper[isotope] = value
                else:
                    helper[isotope] = 0.0
        #
        return helper
    #
    def get_concentration_single(self):
        helper = {}
        #
        if self.var_filetype == "STD":
            for isotope in self.container_lists["ISOTOPES"]:
                concentration_i = self.srm_isotopes[isotope]["Concentration"]
                #
                helper[isotope] = concentration_i
        elif self.var_filetype == "SMPL":
            # concentration_is = float(self.container_files["SMPL"][self.filename_single]["IS Concentration"].get())
            index_file = self.container_lists["SMPL"]["Short"].index(self.filename_single)
            file_long = self.container_lists["SMPL"]["Long"][index_file]
            concentration_is = float(self.container_var["SMPL"][file_long]["IS Data"]["Concentration"].get())
            for isotope in self.container_lists["ISOTOPES"]:
                intensity_ratio_i = self.results_single["Intensity Ratio"][isotope]
                sensitivity_i = self.results_single["Sensitivity"][isotope]
                #
                if sensitivity_i > 0:
                    concentration_i = intensity_ratio_i*(concentration_is/sensitivity_i)
                else:
                    concentration_i = 0.0
                #
                helper[isotope] = concentration_i
        #
        return helper
    #
    def get_lod_single(self):
        helper = {}
        #
        for isotope in self.container_lists["ISOTOPES"]:
            helper[isotope] = []
            #
            n_bg = len(
                self.container_measurements["SELECTED"][self.filename_single][self.var_datatype][isotope]["BG"])
            n_sig = len(
                self.container_measurements["SELECTED"][self.filename_single][self.var_datatype][isotope]["SIG"])
            sensitivity_i = self.results_single["Sensitivity"][isotope]
            dwell_time_i = float(self.container_var["dwell_times"]["Entry"][isotope].get())
            mean_bg_i = np.mean(
                self.container_measurements["SELECTED"][self.filename_single][self.var_datatype][isotope]["BG"])
            error_bg_i = np.std(
                self.container_measurements["SELECTED"][self.filename_single][self.var_datatype][isotope]["BG"], ddof=1)
            #
            if self.filename_single in self.container_lists["STD"]["Short"]:
                var_srm = self.srm_isotopes[isotope]["SRM"]
                var_is = self.container_files[self.var_filetype][self.filename_single]["IS"].get()
                key_element = re.search("(\D+)(\d+)", var_is)
                var_is_element = key_element.group(1)
                concentration_is = self.container_lists["SRM Data"][var_srm][var_is_element]
            else:
                index_file = self.container_lists["SMPL"]["Short"].index(self.filename_single)
                file_long = self.container_lists["SMPL"]["Long"][index_file]
                concentration_is = float(self.container_var["SMPL"][file_long]["IS Data"]["Concentration"].get())
                var_is = self.container_var["SMPL"][file_long]["IS Data"]["IS"].get()
            #
            intensity_is = np.mean(
                self.container_measurements["SELECTED"][self.filename_single][self.var_datatype][var_is]["SIG"]) - \
                           np.mean(
                self.container_measurements["SELECTED"][self.filename_single][self.var_datatype][var_is]["BG"])
            #
            if sensitivity_i > 0:
                if self.container_var["General Settings"]["LOD Selection"].get() == 0:
                    # Pettke et al. (2012)
                    lod_value = (3.29*(mean_bg_i*dwell_time_i*n_sig*(1 + n_sig/n_bg))**0.5 + 2.71)/(
                            n_sig*dwell_time_i*sensitivity_i)*(concentration_is/intensity_is)
                elif self.container_var["General Settings"]["LOD Selection"].get() == 1:
                    # Longerich et al. (1996)
                    lod_value = (3*error_bg_i)/(sensitivity_i)*(concentration_is/intensity_is)*(
                            1/n_bg + 1/n_sig)**0.5
                #
            else:
                lod_value = 0.0
            #
            helper[isotope] = lod_value
        #
        return helper
    #
    def get_intensity_data(self):
        var_id = self.container_var["ma_datareduction"]["Option ID"].get()
        helper = {}
        filetypes = ["STD", "SMPL"]
        categories = ["BACKGROUND", "BACKGROUND ERROR", "SIGNAL", "SIGNAL CORRECTED"]
        n_digits = int(self.container_var["General Settings"]["Accuracy Intensity"].get())
        #
        for filetype in filetypes:
            helper[filetype] = {}
            for category in categories:
                helper[filetype][category] = {}
        #
        for filetype in filetypes:
            #
            if filetype == "STD":
                list_files = self.container_lists[filetype]["Short"]
            elif filetype == "SMPL":
                if var_id != "Select Assemblage":
                    list_files = list(self.container_results[filetype][var_id].keys())
                    list_files.remove("ALL")
                else:
                    list_files = self.container_lists[filetype]["Short"]
            #
            for index, filename in enumerate(list_files):
                cb_file = self.container_var[filetype][self.container_lists[filetype]["Long"][index]]["Checkbox"].get()
                if isinstance(cb_file, str) == True:
                    cb_file = int(cb_file)
                if cb_file == 1:
                    helper[filetype]["BACKGROUND"][filename] = {}
                    helper[filetype]["BACKGROUND ERROR"][filename] = {}
                    helper[filetype]["SIGNAL"][filename] = {}
                    helper[filetype]["SIGNAL CORRECTED"][filename] = {}
                    for isotope in self.container_lists["ISOTOPES"]:
                        var_srm_isotope = self.srm_isotopes[isotope]["SRM"]
                        #
                        if filetype == "STD":
                            var_srm_file = self.container_files["STD"][filename]["SRM"].get()
                            var_is = self.container_files[filetype][filename]["IS"].get()
                            #
                            if var_srm_file == var_srm_isotope and filetype == "STD":
                                intensity_i_bg = np.mean(
                                    self.container_measurements["SELECTED"][filename][self.var_datatype][isotope]["BG"])
                                intensity_i_bg_error = np.std(
                                    self.container_measurements["SELECTED"][filename][self.var_datatype][isotope]["BG"],
                                    ddof=1)
                                intensity_i_mat = np.mean(
                                    self.container_measurements["SELECTED"][filename][self.var_datatype][isotope][
                                        "SIG"])
                                #
                                if np.mean(intensity_i_mat) >= np.mean(intensity_i_bg):
                                    intensity_i = intensity_i_mat - np.mean(intensity_i_bg)
                                else:
                                    intensity_i = 0.0
                                #
                            elif var_srm_file != var_srm_isotope and isotope == var_is:
                                intensity_i_bg = np.mean(
                                    self.container_measurements["SELECTED"][filename][self.var_datatype][isotope]["BG"])
                                intensity_i_bg_error = np.std(
                                    self.container_measurements["SELECTED"][filename][self.var_datatype][isotope]["BG"],
                                    ddof=1)
                                intensity_i_mat = np.mean(
                                    self.container_measurements["SELECTED"][filename][self.var_datatype][isotope][
                                        "SIG"])
                                #
                                if np.mean(intensity_i_mat) >= np.mean(intensity_i_bg):
                                    intensity_i = intensity_i_mat - np.mean(intensity_i_bg)
                                else:
                                    intensity_i = 0.0
                                #
                            else:
                                intensity_i_bg = np.nan
                                intensity_i_bg_error = np.nan
                                intensity_i_mat = np.nan
                                intensity_i = np.nan
                            #
                        else:   # SMPL
                            intensity_i_bg = np.mean(
                                self.container_measurements["SELECTED"][filename][self.var_datatype][isotope]["BG"])
                            intensity_i_bg_error = np.std(
                                self.container_measurements["SELECTED"][filename][self.var_datatype][isotope]["BG"],
                                ddof=1)
                            intensity_i_mat = np.mean(
                                self.container_measurements["SELECTED"][filename][self.var_datatype][isotope]["SIG"])
                            #
                            if np.mean(intensity_i_mat) >= np.mean(intensity_i_bg):
                                intensity_i = intensity_i_mat - np.mean(intensity_i_bg)
                            else:
                                intensity_i = 0.0
                        #
                        helper[filetype]["BACKGROUND"][filename][isotope] = round(intensity_i_bg, n_digits)
                        helper[filetype]["BACKGROUND ERROR"][filename][isotope] = round(intensity_i_bg_error, n_digits)
                        helper[filetype]["SIGNAL"][filename][isotope] = round(intensity_i_mat, n_digits)
                        helper[filetype]["SIGNAL CORRECTED"][filename][isotope] = round(intensity_i, n_digits)
                        #
                        if isotope not in helper[filetype]["BACKGROUND"]:
                            helper[filetype]["BACKGROUND"][isotope] = []
                            helper[filetype]["BACKGROUND ERROR"][isotope] = []
                            helper[filetype]["SIGNAL"][isotope] = []
                            helper[filetype]["SIGNAL CORRECTED"][isotope] = []
                        #
                        intensity_i_bg = round(intensity_i_bg, n_digits)
                        intensity_i_bg_error = round(intensity_i_bg_error, n_digits)
                        intensity_i_mat = round(intensity_i_mat, n_digits)
                        intensity_i = round(intensity_i, n_digits)
                        #
                        if filetype == "STD":
                            if var_srm_file == var_srm_isotope:
                                helper[filetype]["BACKGROUND"][isotope].append(intensity_i_bg)
                                helper[filetype]["BACKGROUND ERROR"][isotope].append(intensity_i_bg_error)
                                helper[filetype]["SIGNAL"][isotope].append(intensity_i_mat)
                                helper[filetype]["SIGNAL CORRECTED"][isotope].append(intensity_i)
                        else:
                            helper[filetype]["BACKGROUND"][isotope].append(intensity_i_bg)
                            helper[filetype]["BACKGROUND ERROR"][isotope].append(intensity_i_bg_error)
                            helper[filetype]["SIGNAL"][isotope].append(intensity_i_mat)
                            helper[filetype]["SIGNAL CORRECTED"][isotope].append(intensity_i)
        #
        ## TESTING
        # print("TESTING: Intensity")
        # print("Filetype: Standard")
        # for file_std in self.container_lists["STD"]["Short"]:
        #     print("File:", file_std)
        #     for isotope in self.container_lists["ISOTOPES"]:
        #         print("Isotope:", isotope)
        #         print("BACKGROUND:", round(helper["STD"]["BACKGROUND"][file_std][isotope], 9))
        #         print("SIGNAL:", round(helper["STD"]["SIGNAL"][file_std][isotope], 9))
        #         print("SIGNAL CORRECTED:", round(helper["STD"]["SIGNAL CORRECTED"][file_std][isotope], 9))
        # print("")
        # print("Filetype: Sample")
        # for file_smpl in self.container_lists["SMPL"]["Short"]:
        #     print("File:", file_smpl)
        #     for isotope in self.container_lists["ISOTOPES"]:
        #         print("Isotope:", isotope)
        #         print("BACKGROUND:", round(helper["SMPL"]["BACKGROUND"][file_smpl][isotope], 9))
        #         print("Data BG:", helper["SMPL"]["BACKGROUND"][file_smpl][isotope])
        #         print("SIGNAL:", round(helper["SMPL"]["SIGNAL"][file_smpl][isotope], 9))
        #         print("Data SIG:", helper["SMPL"]["SIGNAL"][file_smpl][isotope])
        #         print("SIGNAL CORRECTED:", round(helper["SMPL"]["SIGNAL CORRECTED"][file_smpl][isotope], 9))
        #         print("Data SIG CORR:", helper["SMPL"]["SIGNAL CORRECTED"][file_smpl][isotope])
        # for filetype in filetypes:
        #     print(filetype)
        #     for isotope in self.container_lists["ISOTOPES"]:
        #         print(isotope)
        #         print("Background:",
        #               "Mean:", np.nanmean(helper[filetype]["BACKGROUND"][isotope]),
        #               "Error:", np.nanstd(helper[filetype]["BACKGROUND"][isotope], ddof=1),
        #               "Size:", len(helper[filetype]["BACKGROUND"][isotope]))
        #         print("Data BG:", helper[filetype]["BACKGROUND"][isotope])
        #         print("Signal:",
        #               "Mean:", np.nanmean(helper[filetype]["SIGNAL"][isotope]),
        #               "Error:", np.nanstd(helper[filetype]["SIGNAL"][isotope], ddof=1),
        #               "Size:", len(helper[filetype]["SIGNAL"][isotope]))
        #         print("Data SIG:", helper[filetype]["SIGNAL"][isotope])
        #         print("Signal (BG corrected):",
        #               "Mean:", np.nanmean(helper[filetype]["SIGNAL CORRECTED"][isotope]),
        #               "Error:", np.nanstd(helper[filetype]["SIGNAL CORRECTED"][isotope], ddof=1),
        #               "Size:", len(helper[filetype]["SIGNAL CORRECTED"][isotope]))
        #         print("Data SIG CORRECTED:", helper[filetype]["SIGNAL CORRECTED"][isotope])
        #     print("")
        #
        return helper
    #
    def get_intensity_ratio_data(self, intensity):
        helper = {}
        helper["BACKGROUND"] = {}
        helper["SIGNAL"] = {}
        helper["SIGNAL CORRECTED"] = {}
        n_digits = int(self.container_var["General Settings"]["Accuracy Intensity"].get())
        var_id = self.container_var["ma_datareduction"]["Option ID"].get()
        #
        if self.var_filetype == "STD":
            list_files = self.container_lists[self.var_filetype]["Short"]
        elif self.var_filetype == "SMPL":
            if var_id != "Select Assemblage":
                list_files = list(self.container_results[self.var_filetype][var_id].keys())
                list_files.remove("ALL")
            else:
                list_files = self.container_lists[self.var_filetype]["Short"]
        #
        for index, filename in enumerate(list_files):
            cb_file = self.container_var[self.var_filetype][self.container_lists[self.var_filetype]["Long"][index]][
                "Checkbox"].get()
            if isinstance(cb_file, str) == True:
                cb_file = int(cb_file)
            if cb_file == 1:
                helper["BACKGROUND"][filename] = {}
                helper["SIGNAL"][filename] = {}
                helper["SIGNAL CORRECTED"][filename] = {}
                var_is = self.container_files[self.var_filetype][filename]["IS"].get()
                #
                intensity_is_bg = intensity[self.var_filetype]["BACKGROUND"][filename][var_is]
                intensity_is_mat = intensity[self.var_filetype]["SIGNAL"][filename][var_is]
                intensity_is = intensity[self.var_filetype]["SIGNAL CORRECTED"][filename][var_is]
                #
                for isotope in self.container_lists["ISOTOPES"]:
                    #
                    if isotope not in helper["BACKGROUND"]:
                        helper["BACKGROUND"][isotope] = []
                        helper["SIGNAL"][isotope] = []
                        helper["SIGNAL CORRECTED"][isotope] = []
                    #
                    intensity_i_bg = intensity[self.var_filetype]["BACKGROUND"][filename][isotope]
                    intensity_i_mat = intensity[self.var_filetype]["SIGNAL"][filename][isotope]
                    intensity_i = intensity[self.var_filetype]["SIGNAL CORRECTED"][filename][isotope]
                    #
                    if math.isnan(intensity_i_bg) == False:
                        ratio_bg = round(intensity_i_bg/intensity_is_bg, int(1.5*n_digits))
                        ratio_sig = round(intensity_i_mat/intensity_is_mat, int(1.5*n_digits))
                        ratio_sigcorr = round(intensity_i/intensity_is, int(1.5*n_digits))
                    else:
                        ratio_bg = np.nan
                        ratio_sig = np.nan
                        ratio_sigcorr = np.nan
                    #
                    helper["BACKGROUND"][filename][isotope] = ratio_bg
                    helper["SIGNAL"][filename][isotope] = ratio_sig
                    helper["SIGNAL CORRECTED"][filename][isotope] = ratio_sigcorr
                    #
                    helper["BACKGROUND"][isotope].append(ratio_bg)
                    helper["SIGNAL"][isotope].append(ratio_sig)
                    helper["SIGNAL CORRECTED"][isotope].append(ratio_sigcorr)
        #
        ## TESTING
        # print("TESTING: Intensity Ratio")
        # for isotope in self.container_lists["ISOTOPES"]:
        #     print(isotope+"/"+self.var_is)
        #     print("Background:", "Mean:", np.mean(helper["BACKGROUND"][isotope]),
        #           np.std(helper["BACKGROUND"][isotope], ddof=1))
        #     print("Signal:", "Mean:", np.mean(helper["SIGNAL"][isotope]),
        #           np.std(helper["SIGNAL"][isotope], ddof=1))
        #     print("Signal (BG corrected):", "Mean:", np.mean(helper["SIGNAL CORRECTED"][isotope]),
        #           np.std(helper["SIGNAL CORRECTED"][isotope], ddof=1))
        #
        return helper
    #
    def get_sensitivity_data(self, intensity):
        helper = {}
        var_id = self.container_var["ma_datareduction"]["Option ID"].get()
        #
        if self.var_filetype == "STD":
            list_files = self.container_lists[self.var_filetype]["Short"]
        elif self.var_filetype == "SMPL":
            if var_id != "Select Assemblage":
                list_files = list(self.container_results[self.var_filetype][var_id].keys())
                list_files.remove("ALL")
            else:
                list_files = self.container_lists[self.var_filetype]["Short"]
        #
        if self.var_filetype == "STD":
            #
            self.extract_data_times()
            #
            helper["Times"] = {}
            helper["Times"]["STD"] = self.std_times
            helper["Times"]["SMPL"] = self.smpl_times
            #
            for index, filename in enumerate(list_files):
                cb_file = self.container_var["STD"][self.container_lists["STD"]["Long"][index]][
                    "Checkbox"].get()
                if isinstance(cb_file, str) == True:
                    cb_file = int(cb_file)
                if cb_file == 1:
                    helper[filename] = {}
                    #
                    var_srm = self.container_files["STD"][filename]["SRM"].get()
                    list_element_srm = list(self.container_lists["SRM Data"][var_srm].keys())
                    var_is = self.container_files["STD"][filename]["IS"].get()
                    var_srm_is = self.container_files["SRM"][var_is].get()
                    #
                    key_element = re.search("(\D+)(\d+)", var_is)
                    var_is_element = key_element.group(1)
                    #
                    concentration_is = self.srm_data[var_srm][var_is_element]
                    concentration_is = self.container_lists["SRM Data"][var_srm][var_is_element]
                    self.concentration_is_std = concentration_is
                    intensity_is = intensity["STD"]["SIGNAL CORRECTED"][filename][var_is]
                    #
                    for isotope in self.container_lists["ISOTOPES"]:
                        if isotope not in helper:
                                helper[isotope] = []
                        #
                        var_srm_i = self.container_files["SRM"][isotope].get()
                        #
                        if var_srm_i == var_srm:
                            key_element = re.search("(\D+)(\d+)", isotope)
                            var_i_element = key_element.group(1)
                            #
                            var_srm_is = self.container_files["SRM"][var_is].get()
                            #
                            concentration_is = self.container_lists["SRM Data"][var_srm_i][var_is_element]
                            #
                            if var_i_element in self.container_lists["SRM Data"][var_srm_i]:
                                concentration_i = self.container_lists["SRM Data"][var_srm_i][var_i_element]
                            else:
                                concentration_i = 0.0
                            #
                            intensity_i = intensity["STD"]["SIGNAL CORRECTED"][filename][isotope]
                            #
                            intensity_ratio = intensity_i/intensity_is
                            #
                            if var_i_element in self.container_lists["SRM Data"][var_srm_i]:
                                concentration_ratio = concentration_is/concentration_i
                            else:
                                concentration_ratio = 0.0
                            #
                            if math.isnan(intensity_i) == False:
                                sensitivity_i = intensity_ratio*concentration_ratio
                            else:
                                sensitivity_i = np.nan
                            #
                        else:
                            if isotope == var_is:
                                sensitivity_i = 1.0
                            else:
                                sensitivity_i = np.nan
                        #
                        helper[filename][isotope] = sensitivity_i
                        helper[isotope].append(sensitivity_i)
            #
        elif self.var_filetype == "SMPL":
            helper["Optimized"] = {}
            helper["Times"] = {}
            helper["Drift Change"] = {}
            helper_sensitivity_std = {}
            self.xi_opt = {}
            #
            self.extract_data_times()
            #
            helper["Times"]["STD"] = self.std_times
            helper["Times"]["SMPL"] = self.smpl_times
            list_valid_std = []
            list_valid_smpl = []
            #
            for index, file_std in enumerate(self.container_lists["STD"]["Short"]):
                cb_file = self.container_var["STD"][self.container_lists["STD"]["Long"][index]][
                    "Checkbox"].get()
                if isinstance(cb_file, str) == True:
                    cb_file = int(cb_file)
                if cb_file == 1:
                    list_valid_std.append(file_std)
                    #
                    self.xi_std_time[file_std] = {}
                    helper_sensitivity_std[file_std] = {}
                    var_srm = self.container_files["STD"][file_std]["SRM"].get()
                    list_element_srm = list(self.container_lists["SRM Data"][var_srm].keys())
                    var_is = self.container_files["STD"][file_std]["IS"].get()
                    var_srm_is = self.container_files["SRM"][var_is].get()
                    #
                    key_element = re.search("(\D+)(\d+)", var_is)
                    var_is_element = key_element.group(1)
                    #
                    if var_is_element in self.container_lists["SRM Data"][var_srm]:
                        concentration_is = self.container_lists["SRM Data"][var_srm][var_is_element]
                    #
                    intensity_is_bg = np.mean(
                        self.container_measurements["SELECTED"][file_std][self.var_datatype][var_is]["BG"])
                    intensity_is_mat = np.mean(
                        self.container_measurements["SELECTED"][file_std][self.var_datatype][var_is]["SIG"])
                    #
                    if np.mean(intensity_is_mat) >= np.mean(intensity_is_bg):
                        intensity_is = intensity_is_mat - np.mean(intensity_is_bg)
                    else:
                        intensity_is = 0.0
                        # intensity_is = intensity_is_bg - np.mean(intensity_is_mat)
                    #
                    self.concentration_is_std = concentration_is
                    #
                    for isotope, value in self.container_measurements["SELECTED"][file_std][self.var_datatype].items():
                        if isotope in self.isotope_by_srm[var_srm]:
                            key_element = re.search("(\D+)(\d+)", isotope)
                            var_i_element = key_element.group(1)
                            #
                            if var_i_element in list_element_srm:
                                if isotope not in helper_sensitivity_std:
                                    self.xi_opt[isotope] = []
                                    helper_sensitivity_std[isotope] = []
                                    helper_sensitivity_std[file_std][isotope] = {}
                                #
                                intensity_i_bg = np.mean(value["BG"])
                                intensity_i_mat = np.mean(value["SIG"])
                                #
                                if np.mean(intensity_is_mat) >= np.mean(intensity_is_bg):
                                    intensity_i = intensity_i_mat - np.mean(intensity_i_bg)
                                else:
                                    intensity_i = 0.0
                                    # intensity_i = intensity_i_bg - np.mean(intensity_i_mat)
                                #
                                key_element = re.search("(\D+)(\d+)", isotope)
                                var_i_element = key_element.group(1)
                                #
                                var_srm_i = self.container_files["SRM"][isotope].get()
                                var_srm_is = self.container_files["SRM"][var_is].get()
                                #
                                if isotope in self.isotope_by_srm[var_srm_i]:
                                    if var_srm_i != var_srm_is:
                                        concentration_is = self.container_lists["SRM Data"][var_srm_i][var_is_element]
                                        concentration_i = self.container_lists["SRM Data"][var_srm_i][var_i_element]
                                    else:
                                        concentration_i = self.container_lists["SRM Data"][var_srm_is][var_i_element]
                                    #
                                    value = (intensity_i*concentration_is)/(intensity_is*concentration_i)
                                    #
                                    helper_sensitivity_std[isotope].append(value)
                                    helper_sensitivity_std[file_std][isotope] = {
                                        "Time": self.std_times[file_std]["Delta"],
                                        "Sensitivity": value}
                                    #
                                    if isotope not in self.xi_std_time[file_std]:
                                        if self.container_var["General Settings"]["Sensitivity Drift"].get() == 0:
                                            self.xi_std_time[file_std][isotope] = [
                                                self.std_times[file_std]["Delta"], np.mean(value.tolist())]
                                        else:
                                            self.xi_std_time[file_std][isotope] = [
                                                self.std_times[file_std]["Start"], np.mean(value.tolist())]
                            else:
                                if isotope not in helper_sensitivity_std:
                                    self.xi_opt[isotope] = []
                                    helper_sensitivity_std[isotope] = []
                                    helper_sensitivity_std[file_std][isotope] = {}
                                #
                                value = 0.0
                                #
                                helper_sensitivity_std[isotope].append(value)
                                helper_sensitivity_std[file_std][isotope] = {
                                    "Time": self.std_times[file_std]["Delta"],
                                    "Sensitivity": value}
                                #
                                if isotope not in self.xi_std_time[file_std]:
                                    if self.container_var["General Settings"]["Sensitivity Drift"].get() == 0:
                                        self.xi_std_time[file_std][isotope] = [self.std_times[file_std]["Delta"], value]
                                    else:
                                        self.xi_std_time[file_std][isotope] = [self.std_times[file_std]["Start"], value]
            #
            for isotope in self.container_lists["ISOTOPES"]:
                try:
                    helper["Drift Change"][isotope] = (helper_sensitivity_std[isotope][-1]/
                                                       helper_sensitivity_std[isotope][0] - 1)*100
                except:
                    helper["Drift Change"][isotope] = 0.0
            #
            # for key, value in self.xi_std_time.items():
            #     print(key, value)
            for isotope in self.container_lists["ISOTOPES"]:
                self.xi_regr = self.calculate_regression(
                    data=self.xi_std_time, isotope=isotope, file_data=list_valid_std)
                self.xi_opt[isotope].extend(self.xi_regr)
            #
            for index, file_smpl in enumerate(list_files):
                cb_file = self.container_var["SMPL"][self.container_lists["SMPL"]["Long"][index]][
                    "Checkbox"].get()
                if isinstance(cb_file, str) == True:
                    cb_file = int(cb_file)
                if cb_file == 1:
                    list_valid_smpl.append(file_smpl)
                    #
                    helper[file_smpl] = {}
                    for isotope in self.container_lists["ISOTOPES"]:
                        if isotope not in helper:
                            helper[isotope] = []
                        #
                        if self.container_var["General Settings"]["Sensitivity Drift"].get() == 0:
                            value = self.xi_opt[isotope][0]*self.smpl_times[file_smpl]["Delta"] + self.xi_opt[isotope][1]
                        else:
                            value = self.xi_opt[isotope][0]*self.smpl_times[file_smpl]["Start"] + self.xi_opt[isotope][
                                1]
                        #
                        if value >= 0:
                            helper[file_smpl][isotope] = value
                            helper[isotope].append(value)
                            helper["Optimized"][isotope] = self.xi_opt[isotope]
                        else:
                            helper[file_smpl][isotope] = 0.0
                            helper[isotope].append(0.0)
                            helper["Optimized"] = self.xi_opt[isotope]
        #
        ## TESTING
        # print("TESTING: Sensitivity")
        # print("Filetype:", self.var_filetype)
        # for isotope in self.container_lists["ISOTOPES"]:
        #     print(isotope)
        #     print("Dataset size:", len(helper[isotope]))
        #     print("Mean:", np.mean(helper[isotope]), np.std(helper[isotope], ddof=1))
        # for key, value in helper["Times"].items():
        #     print(key, value)
        #
        return helper
    #
    def get_concentration_data(self, intensity, sensitivity):
        helper = {}
        var_id = self.container_var["ma_datareduction"]["Option ID"].get()
        #
        if self.var_filetype == "STD":
            list_files = self.container_lists[self.var_filetype]["Short"]
        elif self.var_filetype == "SMPL":
            if var_id != "Select Assemblage":
                list_files = list(self.container_results[self.var_filetype][var_id].keys())
                list_files.remove("ALL")
            else:
                list_files = self.container_lists[self.var_filetype]["Short"]
        #
        if self.var_filetype == "STD":
            for index, file_std in enumerate(list_files):
                cb_file = self.container_var["STD"][self.container_lists["STD"]["Long"][index]][
                    "Checkbox"].get()
                if isinstance(cb_file, str) == True:
                    cb_file = int(cb_file)
                if cb_file == 1:
                    helper[file_std] = {}
                    var_srm = self.container_files["STD"][file_std]["SRM"].get()
                    #
                    list_element_srm = list(self.container_lists["SRM Data"][var_srm].keys())
                    #
                    for isotope in self.container_lists["ISOTOPES"]:
                        if isotope not in helper:
                            helper[isotope] = []
                        #
                        key_element = re.search("(\D+)(\d+)", isotope)
                        var_i_element = key_element.group(1)
                        #
                        var_srm_i = self.container_files["SRM"][isotope].get()
                        #
                        if var_i_element in list(self.container_lists["SRM Data"][var_srm].keys()):
                            concentration_i = self.container_lists["SRM Data"][var_srm][var_i_element]
                        else:
                            concentration_i = np.nan
                        #
                        if concentration_i >= 0:
                            helper[file_std][isotope] = concentration_i
                            helper[isotope].append(concentration_i)
                        elif concentration_i < 0:
                            helper[file_std][isotope] = 0.0
                            helper[isotope].append(0.0)
                        else:
                            helper[file_std][isotope] = concentration_i
                            helper[isotope].append(concentration_i)
            #
        elif self.var_filetype == "SMPL":
            for index, file_smpl in enumerate(list_files):
                cb_file = self.container_var["SMPL"][self.container_lists["SMPL"]["Long"][index]][
                    "Checkbox"].get()
                if isinstance(cb_file, str) == True:
                    cb_file = int(cb_file)
                if cb_file == 1:
                    helper[file_smpl] = {}
                    #
                    # concentration_is = float(self.container_files["SMPL"][file_smpl]["IS Concentration"].get())
                    concentration_is = float(self.container_var["SMPL"][self.container_lists["SMPL"]["Long"][index]][
                                                 "IS Data"]["Concentration"].get())
                    self.concentration_is_smpl = concentration_is
                    var_is = self.container_files["SMPL"][file_smpl]["IS"].get()
                    var_srm_is = self.container_files["SRM"][var_is].get()
                    intensity_is = intensity[self.var_filetype]["SIGNAL CORRECTED"][file_smpl][var_is]
                    #
                    for isotope in self.container_lists["ISOTOPES"]:
                        var_srm_i = self.container_files["SRM"][isotope].get()
                        if isotope not in helper:
                            helper[isotope] = []
                        #
                        intensity_i = intensity[self.var_filetype]["SIGNAL CORRECTED"][file_smpl][isotope]
                        sensitivity_i = sensitivity[file_smpl][isotope]
                        #
                        if sensitivity_i > 0:
                            concentration_i = (intensity_i/intensity_is)*(concentration_is/sensitivity_i)
                        else:
                            concentration_i = 0.0
                        #
                        if concentration_i >= 0:
                            helper[file_smpl][isotope] = concentration_i
                            helper[isotope].append(concentration_i)
                        else:
                            helper[file_smpl][isotope] = 0.0
                            helper[isotope].append(0.0)
        #
        ## TESTING
        # print("TESTING: Concentration")
        # for isotope in self.container_lists["ISOTOPES"]:
        #     print(isotope)
        #     print("Dataset size:", len(helper[isotope]))
        #     print("Mean:", np.mean(helper[isotope]), np.std(helper[isotope], ddof=1))
        #
        return helper
    #
    def get_rsf_data(self, intensity, sensitivity):
        helper = {}
        var_id = self.container_var["ma_datareduction"]["Option ID"].get()
        #
        if self.var_filetype == "STD":
            list_files = self.container_lists[self.var_filetype]["Short"]
        elif self.var_filetype == "SMPL":
            if var_id != "Select Assemblage":
                list_files = list(self.container_results[self.var_filetype][var_id].keys())
                list_files.remove("ALL")
            else:
                list_files = self.container_lists[self.var_filetype]["Short"]
        #
        if self.var_filetype == "STD":
            for index, file_std in enumerate(list_files):
                cb_file = self.container_var["STD"][self.container_lists["STD"]["Long"][index]][
                    "Checkbox"].get()
                if isinstance(cb_file, str) == True:
                    cb_file = int(cb_file)
                if cb_file == 1:
                    helper[file_std] = {}
                    for isotope in self.container_lists["ISOTOPES"]:
                        if isotope not in helper:
                            helper[isotope] = []
                        #
                        var_srm = self.container_files["STD"][file_std]["SRM"].get()
                        var_srm_isotope = self.srm_isotopes[isotope]["SRM"]
                        #
                        key_element = re.search("(\D+)(\d+)", isotope)
                        var_i_element = key_element.group(1)
                        #
                        key_element_is = re.search("(\D+)(\d+)", self.var_is)
                        var_is_element = key_element_is.group(1)
                        #
                        sensitivity_i = sensitivity[file_std][isotope]
                        intensity_i_std = intensity["STD"]["SIGNAL CORRECTED"][file_std][isotope]
                        intensity_is_std = intensity["STD"]["SIGNAL CORRECTED"][file_std][self.var_is]
                        #
                        if var_srm == var_srm_isotope:
                            if var_i_element in self.container_lists["SRM Data"][var_srm]:
                                concentration_i_std = self.container_lists["SRM Data"][var_srm][var_i_element]
                            else:
                                concentration_i_std = 0.0
                            #
                            concentration_is_std = self.container_lists["SRM Data"][var_srm][var_is_element]
                        else:
                            if var_i_element in self.container_lists["SRM Data"][var_srm_isotope]:
                                concentration_i_std = self.container_lists["SRM Data"][var_srm_isotope][var_i_element]
                            else:
                                concentration_i_std = 0.0
                            #
                            concentration_is_std = self.container_lists["SRM Data"][var_srm_isotope][var_is_element]
                        #
                        if math.isnan(sensitivity_i) == False:
                            try:
                                rsf_value = sensitivity_i*(concentration_i_std)/(intensity_i_std)*(
                                    intensity_is_std)/(concentration_is_std)
                            except:
                                rsf_value = np.nan
                        else:
                            rsf_value = np.nan
                        #
                        helper[file_std][isotope] = rsf_value
                        helper[isotope].append(rsf_value)
            #
        elif self.var_filetype == "SMPL":
            intensity_is_std = np.mean(intensity["STD"]["SIGNAL CORRECTED"][self.var_is])
            intensity_is_smpl = np.mean(intensity["SMPL"]["SIGNAL CORRECTED"][self.var_is])
            #
            for index, file_smpl in enumerate(list_files):
                concentration_is = float(self.container_files["SMPL"][file_smpl]["IS Concentration"].get())
                #
                cb_file = self.container_var["SMPL"][self.container_lists["SMPL"]["Long"][index]][
                    "Checkbox"].get()
                if isinstance(cb_file, str) == True:
                    cb_file = int(cb_file)
                if cb_file == 1:
                    helper[file_smpl] = {}
                    for isotope in self.container_lists["ISOTOPES"]:
                        if isotope not in helper:
                            helper[isotope] = []
                        #
                        intensity_i_std = np.mean(intensity["STD"]["SIGNAL CORRECTED"][isotope])
                        var_srm_isotope = self.srm_isotopes[isotope]["SRM"]
                        #
                        key_element = re.search("(\D+)(\d+)", isotope)
                        var_i_element = key_element.group(1)
                        #
                        if var_i_element in self.container_lists["SRM Data"][var_srm_isotope]:
                            concentration_i = self.container_lists["SRM Data"][var_srm_isotope][var_i_element]
                        else:
                            concentration_i = 0.0
                        #
                        sensitivity_i = sensitivity[file_smpl][isotope]
                        #
                        # rsf_value = (intensity_is_smpl/intensity_is_std)*(
                        #         self.concentration_is_std/self.concentration_is_smpl)
                        rsf_value = sensitivity_i*(concentration_i)/(intensity_i_std)*(intensity_is_smpl)/(concentration_is)
                        #
                        helper[file_smpl][isotope] = rsf_value
                        helper[isotope].append(rsf_value)
        #
        ## TESTING
        # print("TESTING: RSF")
        # for isotope in self.container_lists["ISOTOPES"]:
        #     print(isotope)
        #     print(isotope, helper[isotope])
        #     print("Mean:", np.nanmean(helper[isotope]), np.nanstd(helper[isotope], ddof=1))
        #
        return helper
    #
    def get_lod_data(self, intensity, sensitivity):
        helper = {}
        var_id = self.container_var["ma_datareduction"]["Option ID"].get()
        #
        if self.var_filetype == "STD":
            list_files = self.container_lists[self.var_filetype]["Short"]
        elif self.var_filetype == "SMPL":
            if var_id != "Select Assemblage":
                list_files = list(self.container_results[self.var_filetype][var_id].keys())
                list_files.remove("ALL")
            else:
                list_files = self.container_lists[self.var_filetype]["Short"]
        #
        for index, filename in enumerate(list_files):
            cb_file = self.container_var[self.var_filetype][self.container_lists[self.var_filetype]["Long"][index]][
                "Checkbox"].get()
            if isinstance(cb_file, str) == True:
                cb_file = int(cb_file)
            if cb_file == 1:
                helper[filename] = {}
                #
                for isotope in self.container_lists["ISOTOPES"]:
                    if isotope not in helper:
                            helper[isotope] = []
                    #
                    if isotope in sensitivity[filename]:
                        if self.var_filetype == "STD":  # STANDARD FILES
                            var_srm = self.container_files["STD"][filename]["SRM"].get()
                            key_element = re.search("(\D+)(\d+)", isotope)
                            var_i_element = key_element.group(1)
                            var_srm_i = self.container_files["SRM"][isotope].get()
                            var_is = self.container_files["STD"][filename]["IS"].get()
                            key_element = re.search("(\D+)(\d+)", var_is)
                            var_is_element = key_element.group(1)
                            concentration_is = self.container_lists["SRM Data"][var_srm][var_is_element]
                            intensity_is = intensity[self.var_filetype]["SIGNAL CORRECTED"][filename][var_is]
                            #
                            if var_srm_i == var_srm:
                                mean_bg_i = intensity[self.var_filetype]["BACKGROUND"][filename][isotope]
                                error_bg_i = intensity[self.var_filetype]["BACKGROUND ERROR"][filename][isotope]
                                n_bg = len(self.container_measurements["SELECTED"][filename][self.var_datatype][isotope]["BG"])
                                n_sig = len(self.container_measurements["SELECTED"][filename][self.var_datatype][isotope]["SIG"])
                                sensitivity_i = sensitivity[filename][isotope]
                                dwell_time_i = float(self.container_var["dwell_times"]["Entry"][isotope].get())
                                #
                                if self.container_var["General Settings"]["LOD Selection"].get() == 0:
                                    # Pettke et al. (2012)
                                    if sensitivity_i > 0:
                                        lod_value = (3.29*(mean_bg_i*dwell_time_i*n_sig*(1 + n_sig/n_bg))**0.5 + 2.71)/(
                                                n_sig*dwell_time_i*sensitivity_i)*(concentration_is/intensity_is)
                                    else:
                                        lod_value = 0.0
                                elif self.container_var["General Settings"]["LOD Selection"].get() == 1:
                                    # Longerich et al. (1996)
                                    if sensitivity_i > 0:
                                        lod_value = (3*error_bg_i)/(sensitivity_i)*(concentration_is/intensity_is)*(
                                                1/n_bg + 1/n_sig)**0.5
                                    else:
                                        lod_value = 0.0
                                #
                            else:
                                lod_value = np.nan
                            #
                        else:   # SAMPLE FILES
                            var_is = self.container_files["SMPL"][filename]["IS"].get()
                            # concentration_is = float(self.container_files["SMPL"][filename]["IS Concentration"].get())
                            concentration_is = float(self.container_var["SMPL"][self.container_lists["SMPL"]["Long"][
                                index]]["IS Data"]["Concentration"].get())
                            intensity_is = intensity[self.var_filetype]["SIGNAL CORRECTED"][filename][var_is]
                            #
                            mean_bg_i = intensity[self.var_filetype]["BACKGROUND"][filename][isotope]
                            error_bg_i = intensity[self.var_filetype]["BACKGROUND ERROR"][filename][isotope]
                            n_bg = len(self.container_measurements["SELECTED"][filename][self.var_datatype][isotope]["BG"])
                            n_sig = len(self.container_measurements["SELECTED"][filename][self.var_datatype][isotope]["SIG"])
                            sensitivity_i = sensitivity[filename][isotope]
                            dwell_time_i = float(self.container_var["dwell_times"]["Entry"][isotope].get())
                            #
                            if self.container_var["General Settings"]["LOD Selection"].get() == 0:
                                # Pettke et al. (2012)
                                if sensitivity_i > 0:
                                    lod_value = (3.29*(mean_bg_i*dwell_time_i*n_sig*(1 + n_sig/n_bg))**0.5 + 2.71)/(
                                            n_sig*dwell_time_i*sensitivity_i)*(concentration_is/intensity_is)
                                else:
                                    lod_value = 0.0
                                #
                            elif self.container_var["General Settings"]["LOD Selection"].get() == 1:
                                # Longerich et al. (1996)
                                if sensitivity_i > 0:
                                    lod_value = (3*error_bg_i)/(sensitivity_i)*(concentration_is/intensity_is)*(
                                            1/n_bg + 1/n_sig)**0.5
                                else:
                                    lod_value = 0.0
                        #
                    else:
                        lod_value = np.nan
                    #
                    helper[filename][isotope] = lod_value
                    helper[isotope].append(lod_value)
                #
        ## TESTING
        # print("TESTING: LOD")
        # for isotope in self.container_lists["ISOTOPES"]:
        #     print(isotope)
        #     print("Mean:", np.mean(helper[isotope]), "Error:", np.std(helper[isotope], ddof=1))
        #
        return helper
    #
    def get_normalized_sensitivity(self, sensitivity, intensity, concentration):
        helper = {}
        var_id = self.container_var["ma_datareduction"]["Option ID"].get()
        #
        if self.var_filetype == "STD":
            list_files = self.container_lists[self.var_filetype]["Short"]
        elif self.var_filetype == "SMPL":
            if var_id != "Select Assemblage":
                list_files = list(self.container_results[self.var_filetype][var_id].keys())
                list_files.remove("ALL")
            else:
                list_files = self.container_lists[self.var_filetype]["Short"]
        #
        if self.var_filetype == "STD":
            for file_std in list_files:
                if file_std not in helper:
                    helper[file_std] = {}
                #
                var_is = self.container_files["STD"][file_std]["IS"].get()
                #
                for isotope in self.container_lists["ISOTOPES"]:
                    if isotope not in helper:
                        helper[isotope] = []
                        helper[file_std][isotope] = 0
                    #
                    sensitivity_i = sensitivity[file_std][isotope]
                    intensity_i = intensity["STD"]["SIGNAL CORRECTED"][file_std][var_is]
                    concentration_i = concentration[file_std][var_is]
                    #
                    sensitivity_norm_i = sensitivity_i*(intensity_i/concentration_i)
                    #
                    helper[isotope].append(sensitivity_norm_i)
                    helper[file_std][isotope] = sensitivity_norm_i
        #
        elif self.var_filetype == "SMPL":
            for file_smpl in list_files:
                if file_smpl not in helper:
                    helper[file_smpl] = {}
                #
                var_is = self.container_files["SMPL"][file_smpl]["IS"].get()
                #
                for isotope in self.container_lists["ISOTOPES"]:
                    if isotope not in helper:
                        helper[isotope] = []
                        helper[file_smpl][isotope] = 0
                    #
                    sensitivity_i = sensitivity[file_smpl][isotope]
                    intensity_i = intensity["SMPL"]["SIGNAL CORRECTED"][file_smpl][var_is]
                    concentration_i = concentration[file_smpl][var_is]
                    #
                    sensitivity_norm_i = sensitivity_i*(intensity_i/concentration_i)
                    #
                    helper[isotope].append(sensitivity_norm_i)
                    helper[file_smpl][isotope] = sensitivity_norm_i
        #
        return helper
    #
    def calculate_all_parameters(self):
        results = {}
        categories = ["intensity", "intensity ratio", "sensitivity", "concentration", "RSF", "LOD"]
        categories_data = ["RAW", "SMOOTHED"]
        for category in categories:
            results[category] = None
        #
        intensity = self.get_intensity_data()
        intensity_ratio = self.get_intensity_ratio_data(intensity=intensity)
        sensitivity = self.get_sensitivity_data(intensity=intensity)
        concentration = self.get_concentration_data(intensity=intensity, sensitivity=sensitivity)
        rsf = self.get_rsf_data(intensity=intensity, sensitivity=sensitivity)
        lod = self.get_lod_data(intensity=intensity, sensitivity=sensitivity)
        sensitivity_norm = self.get_normalized_sensitivity(sensitivity=sensitivity, intensity=intensity,
                                                           concentration=concentration)
        #
        results["intensity"] = intensity
        results["intensity ratio"] = intensity_ratio
        results["sensitivity"] = sensitivity
        results["concentration"] = concentration
        results["RSF"] = rsf
        results["LOD"] = lod
        results["sensitivity normalized"] = sensitivity_norm
        #
        # results_all = {}
        # results_all["STD"] = {}
        # results_all["SMPL"] = {}
        # for category in categories:
        #     results_all["STD"][category] = {}
        #     for category_data in categories_data:
        #         results_all["STD"][category][category_data] = {}
        #
        return results
    #
    def calculate_all_parameters_single(self, filename, data_section):
        self.filename_single = filename
        self.data_section = data_section
        self.results_single = {}
        categories = ["Intensity", "Intensity Ratio", "Sensitivity", "Concentration", "LOD"]
        #
        for category in categories:
            self.results_single[category] = None
        #
        intensity = self.get_intensity_single()
        self.results_single["Intensity"] = intensity
        intensity_ratio = self.get_intensity_ratio_single()
        self.results_single["Intensity Ratio"] = intensity_ratio
        sensitivity = self.get_sensitivity_single()
        self.results_single["Sensitivity"] = sensitivity
        concentration = self.get_concentration_single()
        self.results_single["Concentration"] = concentration
        lod = self.get_lod_single()
        self.results_single["LOD"] = lod
        #
        return self.results_single
    #
    def extract_data_times(self):
        self.std_times = {}
        dates_0, times_0 = Data(filename=self.container_lists["STD"]["Long"][0]).import_as_list()
        if self.container_var["General Settings"]["Sensitivity Drift"].get() == 0:
            t_start_0 = datetime.timedelta(
                hours=int(times_0[0][0]), minutes=int(times_0[0][1]), seconds=int(times_0[0][2]))
        else:
            t_start_0 = int(times_0[0][0]) + int(times_0[0][1])/60
        #
        for file in self.container_lists["STD"]["Long"]:
            parts = file.split("/")
            self.std_times[parts[-1]] = {}
            dates, times = Data(filename=file).import_as_list()
            if self.container_var["General Settings"]["Sensitivity Drift"].get() == 0:
                t_start = datetime.timedelta(hours=int(times[0][0]), minutes=int(times[0][1]), seconds=int(times[0][2]))
                t_delta_0 = (t_start - t_start_0).total_seconds()
                self.std_times[parts[-1]]["Start"] = t_start.total_seconds()
                self.std_times[parts[-1]]["Delta"] = t_delta_0
            else:
                t_start = int(times[0][0]) + int(times[0][1])/60
                t_delta_0 = t_start - t_start_0
                self.std_times[parts[-1]]["Start"] = round(t_start, 4)
                self.std_times[parts[-1]]["Delta"] = round(t_delta_0, 4)
        #
        self.smpl_times = {}
        for file in self.container_lists["SMPL"]["Long"]:
            parts = file.split("/")
            self.smpl_times[parts[-1]] = {}
            dates, times = Data(filename=file).import_as_list()
            if self.container_var["General Settings"]["Sensitivity Drift"].get() == 0:
                t_start = datetime.timedelta(hours=int(times[0][0]), minutes=int(times[0][1]), seconds=int(times[0][2]))
                t_delta_0 = (t_start - t_start_0).total_seconds()
                self.smpl_times[parts[-1]]["Start"] = t_start.total_seconds()
                self.smpl_times[parts[-1]]["Delta"] = t_delta_0
            else:
                t_start = int(times[0][0]) + int(times[0][1])/60
                t_delta_0 = t_start - t_start_0
                self.smpl_times[parts[-1]]["Start"] = round(t_start, 4)
                self.smpl_times[parts[-1]]["Delta"] = round(t_delta_0, 4)
        #
    #
    def calculate_regression(self, data, isotope, file_data):
        x_data = []
        y_data = []
        for file in file_data:
            if file in data:
                if isotope in data[file]:
                    x_data.append(data[file][isotope][0])
                    y_data.append(data[file][isotope][1])
        #
        A = np.vstack([x_data, np.ones(len(x_data))]).T
        m, c = np.linalg.lstsq(A, y_data, rcond=None)[0]  # m*x + c
        results = [m, c]
        #
        return results


#!/usr/bin/env python
# -*-coding: utf-8 -*-

#-----------------------------------------------

# Name:		fi_method_matrix_only_tracer.py
# Author:	Maximilian A. Beeskow
# Version:	1.0
# Date:		22.08.2022

#-----------------------------------------------

## MODULES
import numpy as np
import re

class Name:
    #
    def __init__(self):
        name = "Matrix-only Tracer"

class Method:
    #
    def __init__(self, container_measurements, container_lists, container_files, var_is, var_datatype,
                 concentration_host_hostonly, concentration_host_is, host_isotope):
        self.container_measurements = container_measurements
        self.container_lists = container_lists
        self.container_files = container_files
        self.var_is = var_is
        self.var_datatype = var_datatype
        self.concentration_host_hostonly = concentration_host_hostonly
        self.concentration_host_is = concentration_host_is
        self.host_isotope = host_isotope
    #
    def calculate_ratio_R(self, intensity, host_isotope):
        helper = []
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            intensity_mix_hostonly = np.mean(intensity["MIX"][file_smpl][host_isotope])
            intensity_host_hostonly = np.mean(intensity["MATRIX CORRECTED"][file_smpl][host_isotope])
            #
            value = intensity_mix_hostonly/intensity_host_hostonly
            #
            helper.append(value)
            #
        ## TESTING
        # print("Testing: Ratio R")
        # print("Mean:", np.mean(helper), "Error:", np.std(helper, ddof=1))
        #
        return helper
    #
    def calculate_ratio_a(self, intensity, sensitivity, host_isotope):
        helper = []
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            intensity_mix_hostonly = np.mean(intensity["MIX"][file_smpl][host_isotope])
            intensity_mix_is = np.mean(intensity["MIX"][file_smpl][self.var_is])
            sensitivity_hostonly = np.mean(sensitivity[host_isotope])
            #
            value = (intensity_mix_hostonly)/(sensitivity_hostonly*intensity_mix_is)
            #
            helper.append(value)
        #
        ## TESTING
        # print("Testing: Ratio a")
        # print("Mean:", np.mean(helper), "Error:", np.std(helper, ddof=1))
        #
        return helper
    #
    def calculate_ratio_x(self, ratio_a, sensitivity, intensity, calculation_via_a=True):
        helper = []
        #
        concentration_host_hostonly = self.concentration_host_hostonly
        concentration_host_is = self.concentration_host_is
        file_smpl = self.container_lists["SMPL"]["Short"][0]
        concentration_incl_is = float(self.container_files["SMPL"][file_smpl]["IS Concentration"].get())
        #
        if calculation_via_a == True:
            for a in ratio_a:
                value = (concentration_host_hostonly - a*concentration_host_is)/(
                        concentration_host_hostonly - a*(concentration_host_is - concentration_incl_is))
                #
                helper.append(value)
        else:
            helper_2 = {}
            helper_2["I(M,IS)"] = []
            helper_2["I(M,HO)"] = []
            sensitivity_ho = np.mean(sensitivity[self.host_isotope])
            for file_smpl in self.container_lists["SMPL"]["Short"]:
                intensity_mix_is = np.mean(intensity["MIX"][file_smpl][self.var_is])
                intensity_mix_ho = np.mean(intensity["MIX"][file_smpl][self.host_isotope])
                #
                helper_2["I(M,IS)"].append(intensity_mix_is)
                helper_2["I(M,HO)"].append(intensity_mix_ho)
            #
            intensity_mix_is = np.mean(helper_2["I(M,IS)"])
            intensity_mix_ho = np.mean(helper_2["I(M,HO)"])
            #
            value = (sensitivity_ho*intensity_mix_is*concentration_host_hostonly
                     - intensity_mix_ho*concentration_host_is)/(
                    sensitivity_ho*intensity_mix_is*concentration_host_hostonly - intensity_mix_ho*(
                    concentration_host_is - concentration_incl_is))
            #
            helper.append(value)
        #
        ## TESTING
        # print("Testing: Ratio x")
        # print("Mean:", np.mean(helper), "Error:", np.std(helper, ddof=1))
        #
        return helper
    #
    def calculate_intensity_smpl(self, intensity_mat, intensity_mixed, host_isotope):
        #
        intensity_ratio_ri = {}
        intensity_sample = {}
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            intensity_ratio_ri[file_smpl] = np.mean(intensity_mixed[file_smpl][host_isotope])\
                                            /np.mean(intensity_mat[file_smpl][host_isotope])
            intensity_sample[file_smpl] = {}
            for isotope in self.container_lists["ISOTOPES"]:
                intensity_sample[file_smpl][isotope] = intensity_mixed[file_smpl][isotope] \
                                                       - intensity_ratio_ri[file_smpl]\
                                                       *np.mean(intensity_mat[file_smpl][isotope])
        #
        ## TESTING
        # print("RI:", intensity_ratio_ri)
        # print("Intensity (SMPL):", intensity_sample)
        #
        return intensity_sample
    #
    def calculate_concentration_smpl(self, intensity_sample, sensitivity):
        #
        concentration_smpl = {}
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            concentration_smpl[file_smpl] = {}
            concentration_is_smpl = float(self.container_files["SMPL"][file_smpl]["IS Concentration"].get())
            for isotope in self.container_lists["ISOTOPES"]:
                if isotope not in concentration_smpl:
                    concentration_smpl[isotope] = []
                concentration_smpl[file_smpl][isotope] = np.mean((intensity_sample[file_smpl][isotope]))\
                                                         /(np.mean(intensity_sample[file_smpl][self.var_is]))\
                                                         *(concentration_is_smpl)/(np.mean(sensitivity[isotope]))
                value = round(concentration_smpl[file_smpl][isotope], 6)
                if value >= 0:
                    concentration_smpl[isotope].append(abs(value))
                else:
                    concentration_smpl[isotope].append(0.0)
        #
        ## TESTING
        # print("Concentration (IS SMPL):", concentration_is_smpl)
        # print("Concentration (SMPL):", concentration_smpl)
        #
        return concentration_smpl
    #
    def calculate_intensity_inclusion_heinrich_2(self, intensity):
        helper = {}
        helper["Result"] = {}
        helper["I(M,HO)"] = []
        helper["I(H,IS)"] = []
        helper["I(M,i)"] = {}
        helper["I(H,i)"] = {}
        #
        internal_standard = self.var_is
        host_isotope = self.host_isotope
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            intensity_mix_ho = np.mean(intensity["MIX"][file_smpl][host_isotope])
            intensity_host_is = np.mean(intensity["MATRIX CORRECTED"][file_smpl][internal_standard])
            #
            helper["I(M,HO)"].append(intensity_mix_ho)
            helper["I(H,IS)"].append(intensity_host_is)
            #
            for isotope in self.container_lists["ISOTOPES"]:
                if isotope not in helper["Result"]:
                    helper["Result"][isotope] = []
                    helper["I(M,i)"][isotope] = []
                    helper["I(H,i)"][isotope] = []
                #
                intensity_mix_i = np.mean(intensity["MIX"][file_smpl][isotope])
                intensity_host_i = np.mean(intensity["MATRIX CORRECTED"][file_smpl][isotope])
                #
                helper["I(M,i)"][isotope].append(intensity_mix_i)
                helper["I(H,i)"][isotope].append(intensity_host_i)
        #
        intensity_mix_ho = np.mean(helper["I(M,HO)"])
        intensity_host_is = np.mean(helper["I(H,IS)"])
        #
        for isotope in self.container_lists["ISOTOPES"]:
            intensity_mix_i = np.mean(helper["I(M,i)"][isotope])
            intensity_host_i = np.mean(helper["I(H,i)"][isotope])
            #
            value = intensity_mix_i - intensity_mix_ho*(intensity_host_i/intensity_host_is)
            #
            helper["Result"][isotope].append(value)
        #
        ## TESTING
        # print("Testing: Inclusion Intensity (Heinrich)")
        # for key, value in helper["Result"].items():
        #     print("Isotope:", key)
        #     print("Mean:", np.mean(value), "Error:", np.std(value, ddof=1))
        #
        return helper["Result"]
    #
    def calculate_concentration_inclusion_heinrich_2(self, intensity_incl, sensitivity):
        helper = {}
        helper["Results"] = {}
        helper["Xi(i)"] = {}
        helper["I(INCL,i)"] = {}
        #
        internal_standard = self.var_is
        file_smpl = self.container_lists["SMPL"]["Short"][0]
        concentration_incl_is = float(self.container_files["SMPL"][file_smpl]["IS Concentration"].get())
        intensity_incl_is = np.mean(intensity_incl[internal_standard])
        #
        for isotope in self.container_lists["ISOTOPES"]:
            if isotope not in helper["Results"]:
                helper["Results"][isotope] = []
                helper["Xi(i)"][isotope] = []
                helper["I(INCL,i)"][isotope] = []
            #
            sensitivity_i = np.mean(sensitivity[isotope])
            intensity_incl_i = np.mean(intensity_incl[isotope])
            #
            helper["Xi(i)"][isotope].append(sensitivity_i)
            helper["I(INCL,i)"][isotope].append(intensity_incl_i)
        #
        for isotope in self.container_lists["ISOTOPES"]:
            sensitivity_i = np.mean(helper["Xi(i)"][isotope])
            intensity_incl_i = np.mean(helper["I(INCL,i)"][isotope])
            #
            value = (intensity_incl_i/intensity_incl_is)*(concentration_incl_is/sensitivity_i)
            #
            helper["Results"][isotope].append(value)
        #
        ## TESTING
        # print("Testing: Inclusion Concentration (Heinrich)")
        # for key, value in helper["Results"].items():
        #     print("Isotope:", key)
        #     print("Mean:", np.mean(value), "Error:", np.std(value, ddof=1))
        #
        return helper["Results"]
    #
    def calculate_intensity_inclusion(self, host_isotope, intensity_mixed, intensity_mat):
        helper = {}
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            intensity_mixed_t = np.mean(intensity_mixed[file_smpl][host_isotope])
            intensity_mat_t = np.mean(intensity_mat[file_smpl][host_isotope])
            for isotope in self.container_lists["ISOTOPES"]:
                if isotope not in helper:
                    helper[isotope] = []
                intensity_mixed_i = np.mean(intensity_mixed[file_smpl][isotope])
                intensity_mat_i = np.mean(intensity_mat[file_smpl][isotope])
                value = intensity_mixed_i - intensity_mixed_t*intensity_mat_i/intensity_mat_t
                helper[isotope].append(value)
        #
        return helper
    #
    def calculate_concentration_inclusion(self, intensity_incl, intensity_mat):
        helper = {}
        key_element_is = re.search("(\D+)(\d+)", self.var_is)
        var_is_element = key_element_is.group(1)
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            concentration_incl_is = float(self.container_files["SMPL"][file_smpl]["IS Concentration"].get())
            intensity_incl_is = np.mean(intensity_incl[self.var_is])
            for file_std in self.container_lists["STD"]["Short"]:
                var_srm = self.container_files["STD"][file_std]["SRM"].get()
                concentration_std_is = self.container_lists["SRM Data"][var_srm][var_is_element]
                intensity_std_is = np.mean(intensity_mat[file_std][self.var_is])
                for isotope in self.container_lists["ISOTOPES"]:
                    if isotope not in helper:
                        helper[isotope] = []
                    key_element_i = re.search("(\D+)(\d+)", isotope)
                    var_i_element = key_element_i.group(1)
                    concentration_std_i = self.container_lists["SRM Data"][var_srm][var_i_element]
                    intensity_std_i = np.mean(intensity_mat[file_std][isotope])
                    intensity_incl_i = np.mean(intensity_incl[isotope])
                    #
                    value = concentration_incl_is*concentration_std_i/concentration_std_is*intensity_std_is/\
                            intensity_std_i*intensity_incl_i/intensity_incl_is
                    helper[isotope].append(value)
        #
        return helper
    #
    def calculate_intensity_ratio(self, intensity_smpl, var_is):
        helper = {}
        #
        for isotope in self.container_lists["ISOTOPES"]:
            if isotope not in helper:
                helper[isotope] = []
            #
            value = np.mean(intensity_smpl[isotope])/np.mean(intensity_smpl[var_is])
            helper[isotope].append(value)
        #
        return helper
    #
    def calculate_intensity_inclusion_heinrich(self, host_isotope, intensity):
        helper = {}
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            intensity_mix_hostonly = np.mean(intensity["MIX"][file_smpl][host_isotope])
            intensity_host_hostonly = np.mean(intensity["MATRIX CORRECTED"][file_smpl][host_isotope])
            for isotope in self.container_lists["ISOTOPES"]:
                if isotope not in helper:
                    helper[isotope] = []
                intensity_mix_i = np.mean(intensity["MIX"][file_smpl][isotope])
                intensity_host_i = np.mean(intensity["MATRIX CORRECTED"][file_smpl][isotope])
                #
                if np.mean(intensity_mix_i) >= np.mean(intensity_mix_hostonly*(intensity_host_i/intensity_host_hostonly)):
                    value = intensity_mix_i - intensity_mix_hostonly*(intensity_host_i/intensity_host_hostonly)
                else:
                    value = intensity_mix_hostonly*(intensity_host_i/intensity_host_hostonly) - intensity_mix_i
                #
                helper[isotope].append(np.mean(value))
        #
        ## TESTING
        # print("TESTING: Intensity Inclusion")
        # for key, value in helper.items():
        #     print(key)
        #     print("Mean:", np.mean(value), "Error:", np.std(value, ddof=1))
        #
        return helper
    #
    def calculate_concentration_heinrich(self, intensity_incl):
        helper = {}
        key_element_is = re.search("(\D+)(\d+)", self.var_is)
        var_is_element = key_element_is.group(1)
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            concentration_incl_is = float(self.container_files["SMPL"][file_smpl]["IS Concentration"].get())
            intensity_incl_is = np.mean(intensity_incl[self.var_is])
            for file_std in self.container_lists["STD"]["Short"]:
                var_srm = self.container_files["STD"][file_std]["SRM"].get()
                concentration_std_is = self.container_lists["SRM Data"][var_srm][var_is_element]
                #
                intensity_std_is_bg = np.mean(
                    self.container_measurements["SELECTED"][file_std][self.var_datatype][self.var_is]["BG"])
                intensity_std_is_mat = np.mean(
                    self.container_measurements["SELECTED"][file_std][self.var_datatype][self.var_is]["MAT"])
                #
                if np.mean(intensity_std_is_mat) >= np.mean(intensity_std_is_bg):
                    intensity_std_is = np.mean(intensity_std_is_mat - np.mean(intensity_std_is_bg))
                else:
                    intensity_std_is = np.mean(intensity_std_is_bg - np.mean(intensity_std_is_mat))
                #
                for isotope in self.container_lists["ISOTOPES"]:
                    if isotope not in helper:
                        helper[isotope] = []
                    #
                    key_element_i = re.search("(\D+)(\d+)", isotope)
                    var_i_element = key_element_i.group(1)
                    concentration_std_i = self.container_lists["SRM Data"][var_srm][var_i_element]
                    #
                    intensity_std_i_bg = np.mean(
                        self.container_measurements["SELECTED"][file_std][self.var_datatype][isotope]["BG"])
                    intensity_std_i_mat = np.mean(
                        self.container_measurements["SELECTED"][file_std][self.var_datatype][isotope]["MAT"])
                    #
                    if np.mean(intensity_std_i_mat) >= np.mean(intensity_std_i_bg):
                        intensity_std_i = np.mean(intensity_std_i_mat - np.mean(intensity_std_i_bg))
                    else:
                        intensity_std_i = np.mean(intensity_std_i_bg - np.mean(intensity_std_i_mat))
                    #
                    intensity_incl_i = np.mean(intensity_incl[isotope])
                    #
                    value = concentration_incl_is*(concentration_std_i/concentration_std_is)*(
                            intensity_incl_i*intensity_std_is)/(intensity_incl_is*intensity_std_i)
                    #
                    helper[isotope].append(value)
        #
        ## TESTING
        # print("TESTING: Concentration Inclusion")
        # for key, value in helper.items():
        #     print(key)
        #     print("Mean:", np.mean(value), "Error:", np.std(value, ddof=1))
        #
        return helper
    #
    def calculate_concentration_mixed_is(self, ratio_x, intensity, host_isotope):
        helper = {}
        helper["Host IS"] = []
        helper["Host Host Only"] = []
        helper["Mix IS"] = []
        helper["Mix Host Only"] = []
        helper["Result"] = []
        #
        concentration_host_is = self.concentration_host_is
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            intensity_host_is = np.mean(intensity["MATRIX CORRECTED"][file_smpl][self.var_is])
            intensity_host_hostonly = np.mean(intensity["MATRIX CORRECTED"][file_smpl][host_isotope])
            intensity_mix_is = np.mean(intensity["MIX"][file_smpl][self.var_is])
            intensity_mix_hostonly = np.mean(intensity["MIX"][file_smpl][host_isotope])
            #
            helper["Host IS"].append(intensity_host_is)
            helper["Host Host Only"].append(intensity_host_hostonly)
            helper["Mix IS"].append(intensity_mix_is)
            helper["Mix Host Only"].append(intensity_mix_hostonly)
        #
        intensity_host_is = np.mean(helper["Host IS"])
        intensity_host_hostonly = np.mean(helper["Host Host Only"])
        intensity_mix_is = np.mean(helper["Mix IS"])
        intensity_mix_hostonly = np.mean(helper["Mix Host Only"])
        #
        for x in ratio_x:
            value = (1 - x)*concentration_host_is*(intensity_mix_is*intensity_host_hostonly)/(
                    intensity_host_is*intensity_mix_hostonly)
            #
            helper["Result"].append(value)
        #
        ## TESTING
        # print("Testing: Concentration C(IS,MIX)")
        # print("Mean:", np.mean(helper["Result"]), "Error:", np.std(helper["Result"], ddof=1))
        #
        return helper["Result"]
    #
    def calculate_concentration_mixed_hostonly(self, ratio_x):
        helper = []
        concentration_host_hostonly = self.concentration_host_hostonly
        #
        for x in ratio_x:
            value = (1 - x)*concentration_host_hostonly
            #
            helper.append(value)
        #
        ## TESTING
        # print("Testing: Concentration C(HO,MIX)")
        # print("Mean:", np.mean(helper), "Error:", np.std(helper, ddof=1))
        #
        return helper
    #
    def calculate_rsf_mix(self, ratio_x, intensity, host_isotope):
        helper = {}
        helper["Result"] = {}
        helper["Result"]["HOST"] = []
        helper["Result"]["MIX"] = []
        helper["C(IS,STD)"] = []
        helper["I(IS,STD)"] = []
        helper["Host IS"] = []
        helper["Host Host Only"] = []
        helper["Mix Host Only"] = []
        #
        key_element_is = re.search("(\D+)(\d+)", self.var_is)
        var_is_element = key_element_is.group(1)
        concentration_host_is = self.concentration_host_is
        #
        for file_std in self.container_lists["STD"]["Short"]:
            var_srm = self.container_files["STD"][file_std]["SRM"].get()
            concentration_std_is = self.container_lists["SRM Data"][var_srm][var_is_element]
            helper["C(IS,STD)"].append(concentration_std_is)
            #
            intensity_std_is_bg = np.mean(
                self.container_measurements["SELECTED"][file_std][self.var_datatype][self.var_is]["BG"])
            intensity_std_is_mat = np.mean(
                self.container_measurements["SELECTED"][file_std][self.var_datatype][self.var_is]["MAT"])
            #
            if np.mean(intensity_std_is_mat) >= np.mean(intensity_std_is_bg):
                intensity_std_is = np.mean(intensity_std_is_mat - np.mean(intensity_std_is_bg))
            else:
                intensity_std_is = np.mean(intensity_std_is_bg - np.mean(intensity_std_is_mat))
            #
            helper["I(IS,STD)"].append(intensity_std_is)
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            intensity_host_is = np.mean(intensity["MATRIX CORRECTED"][file_smpl][self.var_is])
            intensity_host_hostonly = np.mean(intensity["MATRIX CORRECTED"][file_smpl][host_isotope])
            intensity_mix_hostonly = np.mean(intensity["MIX"][file_smpl][host_isotope])
            #
            helper["Host IS"].append(intensity_host_is)
            helper["Host Host Only"].append(intensity_host_hostonly)
            helper["Mix Host Only"].append(intensity_mix_hostonly)
        #
        concentration_std_is = np.mean(helper["C(IS,STD)"])
        intensity_std_is = np.mean(helper["I(IS,STD)"])
        intensity_host_is = np.mean(helper["Host IS"])
        intensity_host_hostonly = np.mean(helper["Host Host Only"])
        intensity_mix_hostonly = np.mean(helper["Mix Host Only"])
        #
        for x in ratio_x:
            if intensity_std_is > 0 and intensity_host_hostonly > 0:
                value_mix = (concentration_std_is/intensity_std_is)*(1/((1 - x)*concentration_host_is))*(
                        (intensity_host_is*intensity_mix_hostonly)/intensity_host_hostonly)
            else:
                value_mix = 0.0
            #
            helper["Result"]["MIX"].append(value_mix)
            #
            if intensity_std_is > 0:
                value_host = (intensity_host_is*concentration_std_is)/(intensity_std_is*concentration_host_is)
            else:
                value_host = 0.0
            #
            helper["Result"]["HOST"].append(value_host)
        #
        ## TESTING
        # print("Testing: RSF Mix")
        # print("Mean:", np.mean(helper["Result"]["MIX"]), "Error:", np.std(helper["Result"]["MIX"], ddof=1))
        # print("Testing: RSF Host")
        # print("Mean:", np.mean(helper["Result"]["HOST"]), "Error:", np.std(helper["Result"]["HOST"], ddof=1))
        #
        return helper["Result"]
    #
    def calculate_concentration_host_sills(self, sensitivity, intensity):
        helper = {}
        helper["Result"] = {}
        helper["I(H,IS)"] = []
        helper["I(H,i)"] = {}
        helper["Xi(i)"] = {}
        concentration_host_is = self.concentration_host_is
        internal_standard = self.var_is
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            intensity_host_is = np.mean(intensity["MATRIX CORRECTED"][file_smpl][internal_standard])
            #
            helper["I(H,IS)"].append(intensity_host_is)
            for isotope in self.container_lists["ISOTOPES"]:
                if isotope not in helper["Result"]:
                    helper["Result"][isotope] = []
                    helper["I(H,i)"][isotope] = []
                    helper["Xi(i)"][isotope] = []
                #
                sensitivity_i = np.mean(sensitivity[isotope])
                intensity_host_i = np.mean(intensity["MATRIX CORRECTED"][file_smpl][isotope])
                #
                helper["Xi(i)"][isotope].append(sensitivity_i)
                helper["I(H,i)"][isotope].append(intensity_host_i)
        #
        intensity_host_is = np.mean(helper["I(H,IS)"])
        #
        for isotope in self.container_lists["ISOTOPES"]:
            sensitivity_i = np.mean(helper["Xi(i)"][isotope])
            intensity_host_i = np.mean(helper["I(H,i)"][isotope])
            value = (1/sensitivity_i)*(intensity_host_i/intensity_host_is)*concentration_host_is
            #
            helper["Result"][isotope].append(value)
        #
        ## TESTING
        # print("Testing: Host Concentration")
        # for isotope, result in helper["Result"].items():
        #     print("Isotope:", isotope)
        #     print("Mean:", round(np.mean(result), 6), "Error:", round(np.std(result, ddof=1), 6))
        #
        return helper["Result"]
    #
    def calculate_concentration_mix_sills(self, sensitivity, intensity, ratio_x):
        helper = {}
        helper["Result"] = {}
        helper["I(H,IS)"] = []
        helper["I(H,HO)"] = []
        helper["I(M,HO)"] = []
        helper["I(M,i)"] = {}
        helper["Xi(i)"] = {}
        internal_standard = self.var_is
        host_isotope = self.host_isotope
        #
        concentration_host_is = self.concentration_host_is
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            intensity_host_is = np.mean(intensity["MATRIX CORRECTED"][file_smpl][internal_standard])
            intensity_host_ho = np.mean(intensity["MATRIX CORRECTED"][file_smpl][host_isotope])
            intensity_mix_ho = np.mean(intensity["MIX"][file_smpl][host_isotope])
            #
            helper["I(H,IS)"].append(intensity_host_is)
            helper["I(H,HO)"].append(intensity_host_ho)
            helper["I(M,HO)"].append(intensity_mix_ho)
            #
            for isotope in self.container_lists["ISOTOPES"]:
                if isotope not in helper["Result"]:
                    helper["Result"][isotope] = []
                    helper["I(M,i)"][isotope] = []
                    helper["Xi(i)"][isotope] = []
                #
                intensity_mix_i = np.mean(intensity["MIX"][file_smpl][isotope])
                sensitivity_i = np.mean(sensitivity[isotope])
                #
                helper["I(M,i)"][isotope].append(intensity_mix_i)
                helper["Xi(i)"][isotope].append(sensitivity_i)
        #
        intensity_host_is = np.mean(helper["I(H,IS)"])
        intensity_host_ho = np.mean(helper["I(H,HO)"])
        intensity_mix_ho = np.mean(helper["I(M,HO)"])
        #
        for x in ratio_x:
            for isotope in self.container_lists["ISOTOPES"]:
                sensitivity_i = np.mean(helper["Xi(i)"][isotope])
                intensity_mix_i = np.mean(helper["I(M,i)"][isotope])
                #
                value = np.mean((1/sensitivity_i)*(intensity_mix_i/intensity_mix_ho)*(
                        intensity_host_ho/intensity_host_is)*(1 - x)*concentration_host_is)
                #
                helper["Result"][isotope].append(value)
        #
        ## TESTING
        # print("Testing: Mix Concentration")
        # for isotope, result in helper["Result"].items():
        #     print("Isotope:", isotope)
        #     print("Mean:", round(np.mean(result), 6), "Error:", round(np.std(result, ddof=1), 6))
        #
        return helper["Result"]
    #
    def calculate_concentration_inclusion_sills(self, ratio_x, intensity, sensitivity):
        helper = {}
        helper["Result"] = {}
        helper["I(H,IS)"] = []
        helper["I(H,HO)"] = []
        helper["I(M,HO)"] = []
        helper["Xi(i)"] = {}
        helper["I(M,i)"] = {}
        helper["I(H,i)"] = {}
        #
        internal_standard = self.var_is
        host_isotope = self.host_isotope
        concentration_host_is = self.concentration_host_is
        #
        for file_smpl in self.container_lists["SMPL"]["Short"]:
            intensity_host_is = np.mean(intensity["MATRIX CORRECTED"][file_smpl][internal_standard])
            intensity_host_ho = np.mean(intensity["MATRIX CORRECTED"][file_smpl][host_isotope])
            intensity_mix_ho = np.mean(intensity["MIX"][file_smpl][host_isotope])
            #
            helper["I(H,IS)"].append(intensity_host_is)
            helper["I(H,HO)"].append(intensity_host_ho)
            helper["I(M,HO)"].append(intensity_mix_ho)
            #
            for isotope in self.container_lists["ISOTOPES"]:
                if isotope not in helper["Result"]:
                    helper["Result"][isotope] = []
                    helper["Xi(i)"][isotope] = []
                    helper["I(M,i)"][isotope] = []
                    helper["I(H,i)"][isotope] = []
                #
                sensitivity_i = np.mean(sensitivity[isotope])
                intensity_mix_i = np.mean(intensity["MIX"][file_smpl][isotope])
                intensity_host_i = np.mean(intensity["MATRIX CORRECTED"][file_smpl][isotope])
                #
                helper["Xi(i)"][isotope].append(sensitivity_i)
                helper["I(M,i)"][isotope].append(intensity_mix_i)
                helper["I(H,i)"][isotope].append(intensity_host_i)
        #
        intensity_host_is = np.mean(helper["I(H,IS)"])
        intensity_mix_ho = np.mean(helper["I(M,HO)"])
        intensity_host_ho = np.mean(helper["I(H,HO)"])
        x = np.mean(ratio_x)
        #
        for isotope in self.container_lists["ISOTOPES"]:
            sensitivity_i = np.mean(helper["Xi(i)"][isotope])
            intensity_mix_i = np.mean(helper["I(M,i)"][isotope])
            intensity_host_i = np.mean(helper["I(H,i)"][isotope])
            #
            value = ((1 - x)/x)*(concentration_host_is/intensity_host_is)*(1/sensitivity_i)*(
                    (intensity_mix_i/intensity_mix_ho)*intensity_host_ho - intensity_host_i)
            #
            helper["Result"][isotope].append(value)
        #
        ## TESTING
        # print("Testing: Inclusion Concentration")
        # for key, value in helper["Result"].items():
        #     print("Isotope:", key)
        #     print("Mean:", round(np.mean(value), 6), "Error:", round(np.std(value, ddof=1), 6))
        #
        return helper["Result"]
    #
    def calculate_concentration_inclusion_sills_alternative(self, ratio_x, concentration_host, concentration_mix):
        helper = {}
        #
        x = np.mean(ratio_x)
        # file_smpl = self.container_lists["SMPL"]["Short"][0]
        # concentration_incl_is = float(self.container_files["SMPL"][file_smpl]["IS Concentration"].get())
        # x_calib = (np.mean(concentration_mix[self.var_is]) - np.mean(concentration_host[self.var_is]))/(concentration_incl_is - np.mean(concentration_host[self.var_is]))
        # x = x_calib
        #
        for isotope in self.container_lists["ISOTOPES"]:
            if isotope not in helper:
                helper[isotope] = []
            #
            concentration_host_i = np.mean(concentration_host[isotope])
            concentration_mix_i = np.mean(concentration_mix[isotope])
            #
            value = (1 - 1/x)*concentration_host_i + (1/x)*concentration_mix_i
            #
            helper[isotope].append(value)
        #
        ## TESTING
        # print("Testing: Inclusion Concentration (Alternative)")
        # for key, value in helper.items():
        #     print("Isotope:", key)
        #     print("Mean:", round(np.mean(value), 6), "Error:", round(np.std(value, ddof=1), 6))
        #
        return helper
    #
    def do_quantification(self, intensity_mat, intensity_mixed, intensity_incl, host_isotope, sensitivity, intensity):
        #
        ## PREREQUISITES
        # print("Intensity (MAT):", intensity_mat)
        # print("Intensity (MIXED):", intensity_mixed)
        # print("Host Isotope:", host_isotope)
        # print("Sensitivity:", sensitivity)
        #
        ## Calculation: Intensity INCL
        intensity_incl = self.calculate_intensity_inclusion_heinrich(host_isotope=host_isotope, intensity=intensity)
        intensity_incl_heinrich = self.calculate_intensity_inclusion_heinrich_2(intensity=intensity)
        #
        ## Calculation: Ratio R
        ratio_R = self.calculate_ratio_R(intensity=intensity, host_isotope=host_isotope)
        #
        ## Calculation: Ratio a
        ratio_a = self.calculate_ratio_a(intensity=intensity, sensitivity=sensitivity, host_isotope=host_isotope)
        #
        ## Calculation: Ratio x
        ratio_x = self.calculate_ratio_x(ratio_a=ratio_a, intensity=intensity, sensitivity=sensitivity,
                                         calculation_via_a=False)
        #
        ###################
        ## CONCENTRATION ##
        ###################
        #
        ## Host Concentration:
        concentration_host = self.calculate_concentration_host_sills(sensitivity=sensitivity, intensity=intensity)
        #
        ## Mix Concentration:
        concentration_mix = self.calculate_concentration_mix_sills(
            sensitivity=sensitivity, intensity=intensity, ratio_x=ratio_x)
        #
        ## Calculation: C(IS,MIX)
        concentration_mix_is = self.calculate_concentration_mixed_is(ratio_x=ratio_x, intensity=intensity,
                                                                     host_isotope=host_isotope)
        #
        ## Calculation: C(HO,MIX)
        concentration_mix_ho = self.calculate_concentration_mixed_hostonly(ratio_x=ratio_x)
        #
        ## Calculation: RSF(MIX)
        rsf_mix = self.calculate_rsf_mix(ratio_x=ratio_x, intensity=intensity, host_isotope=host_isotope)
        #
        # Calculation: Intensity SMPL
        intensity_smpl = self.calculate_intensity_smpl(intensity_mat=intensity_mat, intensity_mixed=intensity_mixed,
                                                       host_isotope=host_isotope)
        intensity_smpl_2 = self.calculate_intensity_inclusion(
            host_isotope=host_isotope, intensity_mixed=intensity_mixed, intensity_mat=intensity_mat)
        #
        ## Calculation: Intensity Ratio SMPL
        intensity_ratio = self.calculate_intensity_ratio(intensity_smpl=intensity_incl, var_is=self.var_is)
        #
        ## Calculation: Concentration SMPL
        concentration_smpl = self.calculate_concentration_smpl(intensity_sample=intensity_smpl, sensitivity=sensitivity)
        concentration_smpl2 = self.calculate_concentration_inclusion(
            intensity_incl=intensity_smpl_2, intensity_mat=intensity_mat)
        concentration_heinrich = self.calculate_concentration_heinrich(intensity_incl=intensity_incl)
        concentration_heinrich_2 = self.calculate_concentration_inclusion_heinrich_2(
            intensity_incl=intensity_incl_heinrich, sensitivity=sensitivity)
        concentration_sills = self.calculate_concentration_inclusion_sills(
            ratio_x=ratio_x, intensity=intensity, sensitivity=sensitivity)
        concentration_sills_alternative = self.calculate_concentration_inclusion_sills_alternative(
            ratio_x=ratio_x, concentration_host=concentration_host, concentration_mix=concentration_mix)
        #
        ## TESTING
        # print("Intensity (SMPL):", intensity_smpl)
        # print("Intensity (INCL):", intensity_smpl_2)
        # print("Concentration (SMPL):", concentration_smpl)
        # print("Concentration (INCL):", concentration_smpl2)
        #
        return intensity_ratio, concentration_heinrich
